"""scrapli.driver.core.juniper_junos.base_driver"""
from scrapli.driver.base_network_driver import PrivilegeLevel

PRIVS = {
    "exec": (
        PrivilegeLevel(
            pattern=r"^({\w+:\d}\n){0,1}[a-z0-9.\-_@()/:]{1,63}>\s?$",
            name="exec",
            previous_priv="",
            deescalate="",
            escalate="",
            escalate_auth=False,
            escalate_prompt="",
        )
    ),
    "configuration": (
        PrivilegeLevel(
            pattern=r"^({\w+:\d}\[edit\]\n){0,1}[a-z0-9.\-_@()/:]{1,63}#\s?$",
            name="configuration",
            previous_priv="exec",
            deescalate="exit configuration-mode",
            escalate="configure",
            escalate_auth=False,
            escalate_prompt="",
        )
    ),
    "configuration_exclusive": (
        PrivilegeLevel(
            pattern=r"^({\w+:\d}\[edit\]\n){0,1}[a-z0-9.\-_@()/:]{1,63}#\s?$",
            name="configuration_exclusive",
            previous_priv="exec",
            deescalate="exit configuration-mode",
            escalate="configure exclusive",
            escalate_auth=False,
            escalate_prompt="",
        )
    ),
    "configuration_private": (
        PrivilegeLevel(
            pattern=r"^({\w+:\d}\[edit\]\n){0,1}[a-z0-9.\-_@()/:]{1,63}#\s?$",
            name="configuration_private",
            previous_priv="exec",
            deescalate="exit configuration-mode",
            escalate="configure private",
            escalate_auth=False,
            escalate_prompt="",
        )
    ),
}



"""scrapli.exceptions"""


class ScrapliException(Exception):
    """Base Exception for scrapli"""


class SSHNotFound(ScrapliException):
    """Exception for system transport unable to find ssh executable"""


class ScrapliCommandFailure(ScrapliException):
    """Exception for scrapli command/config failures"""


class ScrapliTimeout(ScrapliException):
    """Exception for any scrapli timeouts"""


class ScrapliKeepaliveFailure(ScrapliException):
    """Exception for scrapli missing keepalives"""


class MissingDependencies(ScrapliException):
    """Exception for any missing (probably optional) dependencies"""


class KeyVerificationFailed(ScrapliException):
    """Exception for scrapli public key verification failure"""


class ScrapliAuthenticationFailed(ScrapliException):
    """Exception for scrapli authentication failure"""


class UnknownPrivLevel(ScrapliException):
    """Exception for encountering unknown privilege level"""


class CouldNotAcquirePrivLevel(ScrapliException):
    """Exception for failure to acquire desired privilege level"""


class TransportPluginError(ScrapliException):
    """Exception for transport plugin loading errors"""


class ConnectionNotOpened(ScrapliException):
    """Exception for failures due to connection not being open"""


class UnsupportedPlatform(ScrapliException):
    """Exception for unsupported platform; i.e. using system transport on windows"""



"""scrapli.driver.core.juniper_junos"""
from scrapli.driver.core.juniper_junos.async_driver import AsyncJunosDriver
from scrapli.driver.core.juniper_junos.driver import JunosDriver

__all__ = (
    "AsyncJunosDriver",
    "JunosDriver",
)



"""scrapli.driver.async_generic_driver"""
from collections import UserList
from typing import TYPE_CHECKING, Any, List, Optional, Tuple, Union

from scrapli.channel import AsyncChannel
from scrapli.decorators import TimeoutModifier
from scrapli.driver.async_driver import AsyncScrape
from scrapli.driver.base_generic_driver import GenericDriverBase
from scrapli.response import Response

if TYPE_CHECKING:
    ScrapliMultiResponse = UserList[Response]  # pylint:  disable=E1136; # pragma:  no cover
else:
    ScrapliMultiResponse = UserList


class AsyncGenericDriver(AsyncScrape, GenericDriverBase):
    def __init__(
        self,
        comms_prompt_pattern: str = r"^\S{0,48}[#>$~@:\]]\s*$",
        comms_ansi: bool = True,
        **kwargs: Any,
    ):
        """
        AsyncGenericDriver Object

        Args:
            comms_prompt_pattern: raw string regex pattern -- preferably use `^` and `$` anchors!
                this is the single most important attribute here! if this does not match a prompt,
                scrapli will not work!
                For this GenericDriver the prompt pattern matches a really wide range of things...
                the general pattern is start of line, match any character 0-32 times, then match a
                prompt termination character from the following character set: `#`, `>`, `$`, `~`,
                `@`, `:`, `]`, finally match any or no whitespace till the end of the line. This
                pattern works on all of the "core" platforms and should work on a wide range of
                other devices, however because it is so broad it may also accidentally match too
                many things and cause issues, so be careful!
            comms_ansi: True/False strip comms_ansi characters from output; in the case of the
                GenericDriver, always strip ansi; this may slow things down but will hopefully help
                prevent issues! Obviously can be overridden if desired.
            **kwargs: keyword args to pass to inherited class(es)

        Returns:
            N/A  # noqa: DAR202

        Raises:
            N/A
        """
        super().__init__(comms_prompt_pattern=comms_prompt_pattern, comms_ansi=comms_ansi, **kwargs)
        self.channel: AsyncChannel

    async def get_prompt(self) -> str:
        """
        Convenience method to get device prompt from Channel

        Args:
            N/A

        Returns:
            str: prompt received from channel.get_prompt

        Raises:
            N/A

        """
        prompt: str = await self.channel.get_prompt()
        return prompt

    @TimeoutModifier()
    async def _send_command_eager(
        self,
        command: str,
        strip_prompt: bool = True,
        failed_when_contains: Optional[Union[str, List[str]]] = None,
        *,
        timeout_ops: Optional[float] = None,
    ) -> Response:
        """
        Send a command

        Keeping as a private method rather than adding an argument to the "normal" `send_command`
        method as this should not be used my users generally.

        Args:
            command: string to send to device in privilege exec mode
            strip_prompt: True/False strip prompt from returned output
            failed_when_contains: string or list of strings indicating failure if found in response
            timeout_ops: timeout ops value for this operation; only sets the timeout_ops value for
                the duration of the operation, value is reset to initial value after operation is
                completed

        Returns:
            Response: Scrapli Response object

        Raises:
            N/A

        """
        # decorator cares about timeout_ops, but nothing else does, assign to _ to appease linters
        _ = timeout_ops

        response = self._pre_send_command(
            host=self.transport.host, command=command, failed_when_contains=failed_when_contains
        )
        raw_response, processed_response = await self.channel.send_input(
            channel_input=command, strip_prompt=strip_prompt, eager=True
        )
        return self._post_send_command(
            raw_response=raw_response, processed_response=processed_response, response=response
        )

    @TimeoutModifier()
    async def send_command(
        self,
        command: str,
        strip_prompt: bool = True,
        failed_when_contains: Optional[Union[str, List[str]]] = None,
        *,
        timeout_ops: Optional[float] = None,
    ) -> Response:
        """
        Send a command

        Args:
            command: string to send to device in privilege exec mode
            strip_prompt: True/False strip prompt from returned output
            failed_when_contains: string or list of strings indicating failure if found in response
            timeout_ops: timeout ops value for this operation; only sets the timeout_ops value for
                the duration of the operation, value is reset to initial value after operation is
                completed

        Returns:
            Response: Scrapli Response object

        Raises:
            N/A

        """
        # decorator cares about timeout_ops, but nothing else does, assign to _ to appease linters
        _ = timeout_ops

        response = self._pre_send_command(
            host=self.transport.host, command=command, failed_when_contains=failed_when_contains
        )
        raw_response, processed_response = await self.channel.send_input(
            channel_input=command, strip_prompt=strip_prompt
        )
        return self._post_send_command(
            raw_response=raw_response, processed_response=processed_response, response=response
        )

    async def send_commands(
        self,
        commands: List[str],
        strip_prompt: bool = True,
        failed_when_contains: Optional[Union[str, List[str]]] = None,
        stop_on_failed: bool = False,
        *,
        timeout_ops: Optional[float] = None,
    ) -> ScrapliMultiResponse:
        """
        Send multiple commands

        Args:
            commands: list of strings to send to device in privilege exec mode
            strip_prompt: True/False strip prompt from returned output
            failed_when_contains: string or list of strings indicating failure if found in response
            stop_on_failed: True/False stop executing commands if a command fails, returns results
                as of current execution
            timeout_ops: timeout ops value for this operation; only sets the timeout_ops value for
                the duration of the operation, value is reset to initial value after operation is
                completed. Note that this is the timeout value PER COMMAND sent, not for the total
                of the commands being sent!

        Returns:
            ScrapliMultiResponse: Scrapli MultiResponse object

        Raises:
            N/A

        """
        responses = self._pre_send_commands(commands=commands)
        for command in commands:
            response = await self.send_command(
                command=command,
                strip_prompt=strip_prompt,
                failed_when_contains=failed_when_contains,
                timeout_ops=timeout_ops,
            )
            responses.append(response)
            if stop_on_failed is True and response.failed is True:
                return responses

        return responses

    async def send_commands_from_file(
        self,
        file: str,
        strip_prompt: bool = True,
        failed_when_contains: Optional[Union[str, List[str]]] = None,
        stop_on_failed: bool = False,
        *,
        timeout_ops: Optional[float] = None,
    ) -> ScrapliMultiResponse:
        """
        Send command(s) from file

        Args:
            file: string path to file
            strip_prompt: True/False strip prompt from returned output
            failed_when_contains: string or list of strings indicating failure if found in response
            stop_on_failed: True/False stop executing commands if a command fails, returns results
                as of current execution
            timeout_ops: timeout ops value for this operation; only sets the timeout_ops value for
                the duration of the operation, value is reset to initial value after operation is
                completed. Note that this is the timeout value PER COMMAND sent, not for the total
                of the commands being sent!

        Returns:
            ScrapliMultiResponse: Scrapli MultiResponse object

        Raises:
            N/A

        """
        commands = self._pre_send_commands_from_file(file=file)

        return await self.send_commands(
            commands=commands,
            strip_prompt=strip_prompt,
            failed_when_contains=failed_when_contains,
            stop_on_failed=stop_on_failed,
            timeout_ops=timeout_ops,
        )

    async def send_interactive(
        self,
        interact_events: List[Tuple[str, str, Optional[bool]]],
        failed_when_contains: Optional[Union[str, List[str]]] = None,
        privilege_level: str = "",
    ) -> Response:
        """
        Interact with a device with changing prompts per input.

        Used to interact with devices where prompts change per input, and where inputs may be hidden
        such as in the case of a password input. This can be used to respond to challenges from
        devices such as the confirmation for the command "clear logging" on IOSXE devices for
        example. You may have as many elements in the "interact_events" list as needed, and each
        element of that list should be a tuple of two or three elements. The first element is always
        the input to send as a string, the second should be the expected response as a string, and
        the optional third a bool for whether or not the input is "hidden" (i.e. password input)

        An example where we need this sort of capability:

        ```
        3560CX#copy flash: scp:
        Source filename []? test1.txt
        Address or name of remote host []? 172.31.254.100
        Destination username [carl]?
        Writing test1.txt
        Password:

        Password:
         Sink: C0644 639 test1.txt
        !
        639 bytes copied in 12.066 secs (53 bytes/sec)
        3560CX#
        ```

        To accomplish this we can use the following:

        ```
        interact = conn.channel.send_inputs_interact(
            [
                ("copy flash: scp:", "Source filename []?", False),
                ("test1.txt", "Address or name of remote host []?", False),
                ("172.31.254.100", "Destination username [carl]?", False),
                ("carl", "Password:", False),
                ("super_secure_password", prompt, True),
            ]
        )
        ```

        If we needed to deal with more prompts we could simply continue adding tuples to the list of
        interact "events".

        Args:
            interact_events: list of tuples containing the "interactions" with the device
                each list element must have an input and an expected response, and may have an
                optional bool for the third and final element -- the optional bool specifies if the
                input that is sent to the device is "hidden" (ex: password), if the hidden param is
                not provided it is assumed the input is "normal" (not hidden)
            failed_when_contains: list of strings that, if present in final output, represent a
                failed command/interaction
            privilege_level: ignored in this base class; for LSP reasons for subclasses

        Returns:
            Response: scrapli Response object

        Raises:
            N/A

        """
        _ = privilege_level

        response = self._pre_send_interactive(
            host=self.transport.host,
            interact_events=interact_events,
            failed_when_contains=failed_when_contains,
        )
        raw_response, processed_response = await self.channel.send_inputs_interact(
            interact_events=interact_events
        )
        return self._post_send_command(
            raw_response=raw_response, processed_response=processed_response, response=response
        )

    @TimeoutModifier()
    async def send_and_read(
        self,
        channel_input: str,
        expected_outputs: Optional[List[str]] = None,
        strip_prompt: bool = True,
        failed_when_contains: Optional[Union[str, List[str]]] = None,
        *,
        timeout_ops: Optional[float] = None,
        read_duration: float = 2.5,
    ) -> Response:
        """
        Send an input and read outputs.

        Unlike "normal" scrapli behavior this method reads until the prompt(normal) OR until any of
        a list of expected outputs is seen, OR until the read duration is exceeded. This method does
        not care about/understand privilege levels. This *can* cause you some potential issues if
        not used carefully!

        Args:
            channel_input: input to send to the channel; intentionally named "channel_input" instead
                of "command" or "config" due to this method not caring about privilege levels
            expected_outputs: List of outputs to look for in device response; returns as soon as any
                of the outputs are seen
            strip_prompt: strip prompt or not, defaults to True (yes, strip the prompt)
            failed_when_contains: string or list of strings indicating failure if found in response
            timeout_ops: timeout ops value for this operation; only sets the timeout_ops value for
                the duration of the operation, value is reset to initial value after operation is
                completed
            read_duration:  float duration to read for

        Returns:
            Response: Scrapli Response object

        Raises:
            N/A

        """
        # decorator cares about timeout_ops, but nothing else does, assign to _ to appease linters
        _ = timeout_ops

        response = self._pre_send_command(
            host=self.transport.host,
            command=channel_input,
            failed_when_contains=failed_when_contains,
        )
        raw_response, processed_response = await self.channel.send_input_and_read(
            channel_input=channel_input,
            strip_prompt=strip_prompt,
            expected_outputs=expected_outputs,
            read_duration=read_duration,
        )
        return self._post_send_command(
            raw_response=raw_response, processed_response=processed_response, response=response
        )



def test_configuration_session(nxos_conn):
    nxos_conn.register_configuration_session(session_name="scrapli_test_session")
    result = nxos_conn.send_configs(
        configs=["interface mgmt0", "show configuration session scrapli_test_session"],
        privilege_level="scrapli_test_session",
    )
    nxos_conn.close()
    # pop the config session out
    nxos_conn.privilege_levels.pop("scrapli_test_session")
    assert "config session name scrapli_test_session" in result[1].result


def test_configuration_session_abort(nxos_conn):
    nxos_conn.register_configuration_session(session_name="scrapli_test_session2")
    result = nxos_conn.send_configs(
        configs=["tacocat", "show configuration sessions"],
        privilege_level="scrapli_test_session2",
        stop_on_failed=True,
    )
    current_prompt = nxos_conn.get_prompt()
    nxos_conn.close()
    # pop the config session out
    nxos_conn.privilege_levels.pop("scrapli_test_session2")
    # assert config session was aborted at first sign of failed config
    assert len(result) == 1
    # assert that session aborted and we are back at priv exec
    assert current_prompt == "switch#"



IOSXE_SHOW_VERSION_RAW = """\nCisco IOS XE Software, Version 16.04.01\nCisco IOS Software [Everest], CSR1000V Software (X86_64_LINUX_IOSD-UNIVERSALK9-M), Version 16.4.1, RELEASE SOFTWARE (fc2)\nTechnical Support: http://www.cisco.com/techsupport\nCopyright (c) 1986-2016 by Cisco Systems, Inc.\nCompiled Sun 27-Nov-16 13:02 by mcpre\n\n\nCisco IOS-XE software, Copyright (c) 2005-2016 by cisco Systems, Inc.\nAll rights reserved.  Certain components of Cisco IOS-XE software are\nlicensed under the GNU General Public License ("GPL") Version 2.0.  The\nsoftware code licensed under GPL Version 2.0 is free software that comes\nwith ABSOLUTELY NO WARRANTY.  You can redistribute and/or modify such\nGPL code under the terms of GPL Version 2.0.  For more details, see the\ndocumentation or "License Notice" file accompanying the IOS-XE software,\nor the applicable URL provided on the flyer accompanying the IOS-XE\nsoftware.\n\n\nROM: IOS-XE ROMMON\n\ncsr1000v uptime is 4 hours, 55 minutes\nUptime for this control processor is 4 hours, 56 minutes\nSystem returned to ROM by reload\nSystem image file is "bootflash:packages.conf"\nLast reload reason: reload\n\n\n\nThis product contains cryptographic features and is subject to United\nStates and local country laws governing import, export, transfer and\nuse. Delivery of Cisco cryptographic products does not imply\nthird-party authority to import, export, distribute or use encryption.\nImporters, exporters, distributors and users are responsible for\ncompliance with U.S. and local country laws. By using this product you\nagree to comply with applicable laws and regulations. If you are unable\nto comply with U.S. and local laws, return this product immediately.\n\nA summary of U.S. laws governing Cisco cryptographic products may be found at:\nhttp://www.cisco.com/wwl/export/crypto/tool/stqrg.html\n\nIf you require further assistance please contact us by sending email to\nexport@cisco.com.\n\nLicense Level: ax\nLicense Type: Default. No valid license found.\nNext reload license Level: ax\n\ncisco CSR1000V (VXE) processor (revision VXE) with 2052375K/3075K bytes of memory.\nProcessor board ID 9FKLJWM5EB0\n10 Gigabit Ethernet interfaces\n32768K bytes of non-volatile configuration memory.\n3985132K bytes of physical memory.\n7774207K bytes of virtual hard disk at bootflash:.\n0K bytes of  at webui:.\n\nConfiguration register is 0x2102\n\ncsr1000v#"""
IOSXE_SHOW_VERSION_PROCESSED_NO_STRIP = """Cisco IOS XE Software, Version 16.04.01\nCisco IOS Software [Everest], CSR1000V Software (X86_64_LINUX_IOSD-UNIVERSALK9-M), Version 16.4.1, RELEASE SOFTWARE (fc2)\nTechnical Support: http://www.cisco.com/techsupport\nCopyright (c) 1986-2016 by Cisco Systems, Inc.\nCompiled Sun 27-Nov-16 13:02 by mcpre\n\n\nCisco IOS-XE software, Copyright (c) 2005-2016 by cisco Systems, Inc.\nAll rights reserved.  Certain components of Cisco IOS-XE software are\nlicensed under the GNU General Public License ("GPL") Version 2.0.  The\nsoftware code licensed under GPL Version 2.0 is free software that comes\nwith ABSOLUTELY NO WARRANTY.  You can redistribute and/or modify such\nGPL code under the terms of GPL Version 2.0.  For more details, see the\ndocumentation or "License Notice" file accompanying the IOS-XE software,\nor the applicable URL provided on the flyer accompanying the IOS-XE\nsoftware.\n\n\nROM: IOS-XE ROMMON\n\ncsr1000v uptime is 4 hours, 55 minutes\nUptime for this control processor is 4 hours, 56 minutes\nSystem returned to ROM by reload\nSystem image file is "bootflash:packages.conf"\nLast reload reason: reload\n\n\n\nThis product contains cryptographic features and is subject to United\nStates and local country laws governing import, export, transfer and\nuse. Delivery of Cisco cryptographic products does not imply\nthird-party authority to import, export, distribute or use encryption.\nImporters, exporters, distributors and users are responsible for\ncompliance with U.S. and local country laws. By using this product you\nagree to comply with applicable laws and regulations. If you are unable\nto comply with U.S. and local laws, return this product immediately.\n\nA summary of U.S. laws governing Cisco cryptographic products may be found at:\nhttp://www.cisco.com/wwl/export/crypto/tool/stqrg.html\n\nIf you require further assistance please contact us by sending email to\nexport@cisco.com.\n\nLicense Level: ax\nLicense Type: Default. No valid license found.\nNext reload license Level: ax\n\ncisco CSR1000V (VXE) processor (revision VXE) with 2052375K/3075K bytes of memory.\nProcessor board ID 9FKLJWM5EB0\n10 Gigabit Ethernet interfaces\n32768K bytes of non-volatile configuration memory.\n3985132K bytes of physical memory.\n7774207K bytes of virtual hard disk at bootflash:.\n0K bytes of  at webui:.\n\nConfiguration register is 0x2102\n\ncsr1000v#"""
IOSXE_SHOW_VERSION_PROCESSED_STRIP = IOSXE_SHOW_VERSION_PROCESSED_NO_STRIP[:-11]
IOSXE_EXEC_PROMPT = "csr1000v>"
IOSXE_PRIV_EXEC_PROMPT = "csr1000v#"
IOSXE_CONFIG_PROMPT = "csr1000v(config)#"

IOSXE_TEST_CASES = {
    "test_read_until_prompt": {"expected_prompt": IOSXE_PRIV_EXEC_PROMPT},
    "test_get_prompt": {
        "exec": IOSXE_EXEC_PROMPT,
        "privilege_exec": IOSXE_PRIV_EXEC_PROMPT,
        "configuration": IOSXE_CONFIG_PROMPT,
    },
    "test_send_input": {
        "raw_result": IOSXE_SHOW_VERSION_RAW,
        "processed_result": {
            "strip": IOSXE_SHOW_VERSION_PROCESSED_STRIP,
            "no_strip": IOSXE_SHOW_VERSION_PROCESSED_NO_STRIP,
        },
    },
    "test_send_inputs_interact": {
        "interact_events": [
            ["clear logging", "Clear logging buffer [confirm]"],
            ["", IOSXE_PRIV_EXEC_PROMPT],
        ],
        "raw_result": "clear logging\nClear logging buffer [confirm]\ncsr1000v#",
        "processed_result": "clear logging\nClear logging buffer [confirm]\ncsr1000v#",
    },
}

TEST_CASES = {"cisco_iosxe": IOSXE_TEST_CASES}



"""scrapli.ssh_config"""
import os
import re
import shlex
import sys
from copy import deepcopy
from typing import Dict, Optional

if sys.version_info >= (3, 8):
    Match = re.Match
else:
    from typing import Match  # pragma:  no cover

HOST_ATTRS = (
    "port",
    "user",
    "address_family",
    "bind_address",
    "connect_timeout",
    "identities_only",
    "identity_file",
    "keyboard_interactive",
    "password_authentication",
    "preferred_authentication",
)


class SSHConfig:
    def __init__(self, ssh_config_file: str) -> None:
        """
        Initialize SSHConfig Object

        Parse OpenSSH config file

        Try to load the following data for all entries in config file:
            Host
            HostName
            Port
            User
            *AddressFamily
            *BindAddress
            *ConnectTimeout
            IdentitiesOnly
            IdentityFile
            *KbdInteractiveAuthentication
            *PasswordAuthentication
            *PreferredAuthentications

        * items are mostly ready to load but are unused in scrapli right now so are not being set
        at this point.

        NOTE: this does *not* accept duplicate "*" entries -- the final "*" entry will overwrite any
        previous "*" entries. In general for system transport this shouldn't matter much because
        scrapli only cares about parsing the config file to see if a key (any key) exists for a
        given host (we care about that because ideally we use "pipes" auth, but this is only an
        option if we have a key to auth with).

        Args:
            ssh_config_file: string path to ssh configuration file

        Returns:
            N/A  # noqa: DAR202

        Raises:
            TypeError: if non-string value provided for ssh_config_file

        """
        if not isinstance(ssh_config_file, str):
            raise TypeError(f"`ssh_config_file` expected str, got {type(ssh_config_file)}")

        self.ssh_config_file = os.path.expanduser(ssh_config_file)
        if self.ssh_config_file:
            with open(self.ssh_config_file, "r") as f:
                self.ssh_config = f.read()
            self.hosts = self._parse()
            if not self.hosts:
                self.hosts = {}
            if "*" not in self.hosts.keys():
                self.hosts["*"] = Host()
                self.hosts["*"].hosts = "*"
        else:
            self.hosts = {}
            self.hosts["*"] = Host()
            self.hosts["*"].hosts = "*"

        # finally merge all args from less specific hosts into the more specific hosts, preserving
        # the options from the more specific hosts of course
        self._merge_hosts()

    def __str__(self) -> str:
        """
        Magic str method for SSHConfig class

        Args:
            N/A

        Returns:
            str: string representation of object

        Raises:
            N/A

        """
        return "SSHConfig Object"

    def __repr__(self) -> str:
        """
        Magic repr method for SSHConfig class

        Args:
            N/A

        Returns:
            str: repr for class object

        Raises:
            N/A

        """
        class_dict = self.__dict__.copy()
        del class_dict["ssh_config"]
        return f"SSHConfig {class_dict}"

    def __bool__(self) -> bool:
        """
        Magic bool method; return True if ssh_config_file

        Args:
            N/A

        Returns:
            bool: True/False if ssh_config_file

        Raises:
            N/A

        """
        if self.ssh_config:
            return True
        return False

    @staticmethod
    def _strip_comments(line: str) -> str:
        """
        Strip out comments from ssh config file lines

        Args:
            line: to strip comments from

        Returns:
            str: rejoined ssh config file line after stripping comments

        Raises:
            N/A

        """
        line = " ".join(shlex.split(line, comments=True))
        return line

    def _parse(self) -> Dict[str, "Host"]:
        """
        Parse SSH configuration file

        Args:
            N/A

        Returns:
            discovered_hosts: dict of host objects discovered in ssh config file

        Raises:
            N/A

        """
        # uncomment next line and handle global patterns (stuff before hosts) at some point
        # global_config_pattern = re.compile(r"^.*?\b(?=host)", flags=re.I | re.S)
        # use word boundaries with a positive lookahead to get everything between the word host
        # need to do this as whitespace/formatting is not really a thing in ssh_config file
        # match host\s to ensure we don't pick up hostname and split things there accidentally
        host_pattern = re.compile(r"\bhost.*?\b(?=host\s|\s+$|$)", flags=re.I | re.S)
        host_entries = re.findall(pattern=host_pattern, string=self.ssh_config)

        discovered_hosts: Dict[str, Host] = {}
        if not host_entries:
            return discovered_hosts

        # do we need to add whitespace between match and end of line to ensure we match correctly?
        hosts_pattern = re.compile(r"^\s*host[\s=]+(.*)$", flags=re.I | re.M)
        hostname_pattern = re.compile(r"^\s*hostname[\s=]+([\w.-]*)$", flags=re.I | re.M)
        port_pattern = re.compile(r"^\s*port[\s=]+([\d]*)$", flags=re.I | re.M)
        user_pattern = re.compile(r"^\s*user[\s=]+([\w]*)$", flags=re.I | re.M)
        # address_family_pattern = None
        # bind_address_pattern = None
        # connect_timeout_pattern = None
        identities_only_pattern = re.compile(
            r"^\s*identitiesonly[\s=]+(yes|no)$", flags=re.I | re.M
        )
        identity_file_pattern = re.compile(
            r"^\s*identityfile[\s=]+([\w.\/\@~-]*)$", flags=re.I | re.M
        )
        # keyboard_interactive_pattern = None
        # password_authentication_pattern = None
        # preferred_authentication_pattern = None

        for host_entry in host_entries:
            host = Host()
            host_line = re.search(pattern=hosts_pattern, string=host_entry)
            if isinstance(host_line, Match):
                host.hosts = self._strip_comments(host_line.groups()[0])
            else:
                host.hosts = ""
            hostname = re.search(pattern=hostname_pattern, string=host_entry)
            if isinstance(hostname, Match):
                host.hostname = self._strip_comments(hostname.groups()[0])
            port = re.search(pattern=port_pattern, string=host_entry)
            if isinstance(port, Match):
                host.port = int(self._strip_comments(port.groups()[0]))
            user = re.search(pattern=user_pattern, string=host_entry)
            if isinstance(user, Match):
                host.user = self._strip_comments(user.groups()[0])
            # address_family = re.search(user_pattern, host_entry[0])
            # bind_address = re.search(user_pattern, host_entry[0])
            # connect_timeout = re.search(user_pattern, host_entry[0])
            identities_only = re.search(pattern=identities_only_pattern, string=host_entry)
            if isinstance(identities_only, Match):
                host.identities_only = self._strip_comments(identities_only.groups()[0])
            identity_file = re.search(pattern=identity_file_pattern, string=host_entry)
            if isinstance(identity_file, Match):
                host.identity_file = os.path.expanduser(
                    self._strip_comments(identity_file.groups()[0])
                )
            # keyboard_interactive = re.search(user_pattern, host_entry[0])
            # password_authentication = re.search(user_pattern, host_entry[0])
            # preferred_authentication = re.search(user_pattern, host_entry[0])
            discovered_hosts[host.hosts] = host
        return discovered_hosts

    def _merge_hosts(self) -> None:
        """
        Merge less specific host pattern data into a given host

        Args:
            N/A

        Returns:
            N/A  # noqa: DAR202

        Raises:
            N/A

        """
        for host in self.hosts:
            _current_hosts = deepcopy(self.hosts)
            while True:
                fuzzy_match = self._lookup_fuzzy_match(host=host, hosts=_current_hosts)
                for attr in HOST_ATTRS:
                    if not getattr(self.hosts[host], attr):
                        setattr(self.hosts[host], attr, getattr(self.hosts[fuzzy_match], attr))
                try:
                    _current_hosts.pop(fuzzy_match)
                except KeyError:
                    # this means we hit the "*" entry twice and we can bail out
                    break

    def _lookup_fuzzy_match(self, host: str, hosts: Optional[Dict[str, "Host"]] = None) -> str:
        """
        Look up fuzzy matched hosts

        Get the best match ssh config Host entry for a given host; this allows for using
        the splat and question-mark operators in ssh config file

        Args:
            host: host to lookup in discovered_hosts dict
            hosts: hosts dict to operate on; used for passing in partial dict of hosts while
                performing merge operations

        Returns:
            str: Nearest match (if applicable) host or `*` if none found

        Raises:
            N/A

        """
        hosts = hosts or self.hosts

        possible_matches = []
        for host_entry in hosts.keys():
            host_list = host_entry.split()
            for host_pattern in host_list:
                # replace periods with literal period
                # replace asterisk (match 0 or more things) with appropriate regex
                # replace question mark (match one thing) with appropriate regex
                cleaned_host_pattern = (
                    host_pattern.replace(".", r"\.").replace("*", r"(.*)").replace("?", r"(.)")
                )
                # compile with case insensitive
                search_pattern = re.compile(cleaned_host_pattern, flags=re.I)
                result = re.search(pattern=search_pattern, string=host)
                # if we get a result, append it and the original pattern to the possible matches
                if result:
                    possible_matches.append((result, host_entry))

        # initialize a None best match
        current_match = None
        for match in possible_matches:
            if current_match is None:
                current_match = match
            # count how many chars were replaced to get regex to work
            chars_replaced = 0
            for start_char, end_char in match[0].regs[1:]:
                chars_replaced += end_char - start_char
            # count how many chars were replaced to get regex to work on best match
            best_match_chars_replaced = 0
            for start_char, end_char in current_match[0].regs[1:]:
                best_match_chars_replaced += end_char - start_char
            # if match replaced less chars than "best_match" we have a new best match
            if chars_replaced < best_match_chars_replaced:
                current_match = match
        if current_match is not None:
            best_match = current_match[1]
        else:
            best_match = "*"
        return best_match

    def lookup(self, host: str) -> "Host":
        """
        Lookup a given host

        Args:
            host: host to lookup in discovered_hosts dict

        Returns:
            Host: best matched host from parsed ssh config file hosts, "*" if no better match found

        Raises:
            N/A

        """
        # return exact 1:1 match if exists
        if host in self.hosts.keys():
            return self.hosts[host]
        # return match if given host is an exact match for a host entry
        for host_entry in self.hosts.keys():
            host_list = host_entry.split()
            if host in host_list:
                return self.hosts[host_entry]
        # otherwise need to select the most correct host entry
        fuzzy_match = self._lookup_fuzzy_match(host)
        return self.hosts[fuzzy_match]


class Host:
    def __init__(self) -> None:
        """
        Host Object

        Create a Host object based on ssh config file information
        """
        self.hosts: str = ""
        self.hostname: Optional[str] = None
        self.port: Optional[int] = None
        self.user: str = ""
        self.address_family: Optional[str] = None
        self.bind_address: Optional[str] = None
        self.connect_timeout: Optional[str] = None
        self.identities_only: Optional[str] = None
        self.identity_file: Optional[str] = None
        self.keyboard_interactive: Optional[str] = None
        self.password_authentication: Optional[str] = None
        self.preferred_authentication: Optional[str] = None

    def __str__(self) -> str:
        """
        Magic str method for HostEntry class

        Args:
            N/A

        Returns:
            str: string for class object

        Raises:
            N/A

        """
        return f"Host: {self.hosts}"

    def __repr__(self) -> str:
        """
        Magic repr method for HostEntry class

        Args:
            N/A

        Returns:
            str: repr for class object

        Raises:
            N/A

        """
        class_dict = self.__dict__.copy()
        return f"Host {class_dict}"


class SSHKnownHosts:
    def __init__(self, ssh_known_hosts_file: str) -> None:
        """
        Initialize SSHKnownHosts Object

        Parse OpenSSH known hosts file

        Try to load the following data for all entries in known hosts file:
            Host
            Key Type
            Public Key

        Args:
            ssh_known_hosts_file: string path to ssh known hosts file

        Returns:
            N/A  # noqa: DAR202

        Raises:
            TypeError: if non-string value provided for ssh_known_hosts

        """
        if not isinstance(ssh_known_hosts_file, str):
            raise TypeError(
                f"`ssh_known_hosts_file` expected str, got {type(ssh_known_hosts_file)}"
            )

        self.ssh_known_hosts_file = os.path.expanduser(ssh_known_hosts_file)
        if self.ssh_known_hosts_file:
            with open(self.ssh_known_hosts_file, "r") as f:
                self.ssh_known_hosts = f.read()
            self.hosts = self._parse()
            if not self.hosts:
                self.hosts = {}
        else:
            self.hosts = {}

    def _parse(self) -> Dict[str, Dict[str, str]]:
        """
        Parse SSH configuration file

        Args:
            N/A

        Returns:
            discovered_hosts: dict of host objects discovered in known hosts file

        Raises:
            N/A

        """
        # match any non whitespace from start of the line... this should cover v4/v6/names
        # skip a space and match any word (also w/ hyphen) to get key type, lastly
        # match any non whitespace to the end of the line to get the public key
        host_pattern = re.compile(r"^\S+\s[\w\-]+\s\S+$", flags=re.I | re.M)
        host_entries = re.findall(pattern=host_pattern, string=self.ssh_known_hosts)

        known_hosts: Dict[str, Dict[str, str]] = {}
        for host_entry in host_entries:
            host, key_type, public_key = host_entry.split()
            # to simplify lookups down the line, split any list of hosts and just create a unique
            # entry per host
            for individual_host in host.split(","):
                known_hosts[individual_host] = {}
                known_hosts[individual_host]["key_type"] = key_type
                known_hosts[individual_host]["public_key"] = public_key

        return known_hosts



"""scrapli.decorators"""
import asyncio
import multiprocessing.pool
import signal
import sys
import threading
from typing import TYPE_CHECKING, Any, Callable, Dict, Union

from scrapli.exceptions import ConnectionNotOpened, ScrapliTimeout

if TYPE_CHECKING:
    from scrapli.channel import AsyncChannel  # pragma: no cover
    from scrapli.channel import Channel  # pragma:  no cover
    from scrapli.driver import AsyncGenericDriver, GenericDriver  # pragma:  no cover
    from scrapli.transport import Transport  # pragma:  no cover

WIN = sys.platform.startswith("win")


class OperationTimeout:
    def __init__(self, attribute: str, message: str = "") -> None:
        """
        Operation timeout decorator

        Wrap an operation, check class for given attribute and use that for the timeout duration.

        Historically this was not a class and has gone through several iterations... the first
        iteration was using only signal... this was fast and efficient but did not work on windows
        and due to system transport spawning a pty/forking things it would not work for that either.
        Timeouts were then moved to use the multiprocessing method, which works in all cases, and is
        thread safe (unlike signal), however it is very cpu intensive and slightly slower than the
        signal method. This current iteration moved the decorator into a class so it is more orderly
        and easier to break things up into smaller chunks, and importantly now supports both timeout
        methods -- signal and multiprocessing. This will always try to use the signal method, but if
        that is not available (due to windows, system transport, or being in not the main thread),
        it will fallback to the multiprocessing method.

        Args:
            attribute: name of attribute to use to get timeout value (set in the decorator)
            message: optional message to use in timeout exception (set in the decorator)

        Returns:
            N/A  # noqa: DAR202

        Raises:
            N/A

        """
        self.attribute = attribute
        self.message = message

        self.scrapli_obj: Union["Channel", "Transport"]
        self.session_lock: threading.Lock
        self.close: Callable[..., Any]
        self.transport: "Transport"
        self.timeout_duration: float
        self.timeout_exit: bool = True
        self.signals_supported_transport: bool = True
        self._use_signals: bool = False

    def __call__(self, wrapped_func: Callable[..., Any]) -> Callable[..., Any]:
        """
        Operation timeout decorator

        This is what is called when the decorator is triggered

        Args:
            wrapped_func: function being decorated

        Returns:
            decorate: decorated func

        Raises:
            N/A

        """
        if asyncio.iscoroutinefunction(wrapped_func):

            async def decorate(*args: Any, **kwargs: Any) -> Any:
                self.scrapli_obj = args[0]
                self.timeout_duration = getattr(self.scrapli_obj, self.attribute, None)

                if not self.timeout_duration:
                    self.scrapli_obj.logger.info(
                        f"Could not find {self.attribute} value of {self.scrapli_obj}, continuing "
                        "without timeout decorator"
                    )
                    return await wrapped_func(*args, **kwargs)

                self.set_scrapli_obj_attrs()
                return await self.asyncio_timeout(
                    wrapped_func=wrapped_func, args=args, kwargs=kwargs
                )

        else:

            def decorate(*args: Any, **kwargs: Any) -> Any:  # type: ignore
                self.scrapli_obj = args[0]
                self.timeout_duration = getattr(self.scrapli_obj, self.attribute, None)

                if not self.timeout_duration:
                    self.scrapli_obj.logger.info(
                        f"Could not find {self.attribute} value of {self.scrapli_obj}, continuing "
                        "without timeout decorator"
                    )
                    return wrapped_func(*args, **kwargs)

                self.set_scrapli_obj_attrs()
                self.determine_sync_timeout_method()

                if self._use_signals:
                    return self.signal_timeout(wrapped_func=wrapped_func, args=args, kwargs=kwargs)
                return self.multiprocessing_timeout(
                    wrapped_func=wrapped_func, args=args, kwargs=kwargs
                )

        return decorate

    def set_scrapli_obj_attrs(self) -> None:
        """
        Parent method to set attributes of wrapped functions class object

        Args:
            N/A

        Returns:
            N/A  # noqa: DAR202

        Raises:
            N/A

        """
        from scrapli.channel import AsyncChannel, Channel  # pylint: disable=C0415

        if isinstance(self.scrapli_obj, (AsyncChannel, Channel)):
            self.timeout_exit = self.scrapli_obj.transport.timeout_exit
            self.session_lock = self.scrapli_obj.transport.session_lock
            self.close = self.scrapli_obj.transport.close
            self.transport = self.scrapli_obj.transport
        else:
            self.timeout_exit = self.scrapli_obj.timeout_exit
            self.session_lock = self.scrapli_obj.session_lock
            self.close = self.scrapli_obj.close
            self.transport = self.scrapli_obj

    def determine_sync_timeout_method(self) -> None:
        """
        Decide what timeout mechanism to use for synchronous usage

        Args:
            N/A

        Returns:
            N/A  # noqa: DAR202

        Raises:
            N/A

        """
        from scrapli.transport.systemssh import SystemSSHTransport  # pylint: disable=C0415
        from scrapli.transport.telnet import TelnetTransport  # pylint: disable=C0415

        if isinstance(self.transport, (SystemSSHTransport, TelnetTransport)):
            # system transport cant use signals due to pty, unclear why telnetlib doesnt work with
            # signals, but it does not
            self.signals_supported_transport = False

        if (
            not self.signals_supported_transport
            or WIN
            or threading.current_thread() is not threading.main_thread()
        ):
            self._use_signals = False
        else:
            self._use_signals = True

    def _handle_timeout(self) -> None:
        """
        Timeout handler method to release locks/raise exception consistently between timeout methods

        Args:
            N/A

        Returns:
            N/A  # noqa: DAR202

        Raises:
            ScrapliTimeout: always, if we hit this method we have already timed out!

        """
        from scrapli.channel import AsyncChannel, Channel  # pylint: disable=C0415

        if self.timeout_exit:
            self.scrapli_obj.logger.info("timeout_exit is True, closing transport")
            if self.session_lock.locked():
                self.session_lock.release()
            self.close()
            if not isinstance(self.scrapli_obj, (AsyncChannel, Channel)):
                # if system transport is timing out then we can encounter a condition where timeout
                # happens in system, and we close the transport, however we then still have to deal
                # with unlocking the lock in `_send_input` (for example) which causes a RuntimeError
                # if the lock has been unlocked during closing of the session, so re-acquire the
                # lock if this timeout occurred in NOT a channel object - there needs to be a big
                # overhaul to all the lock handling but that will be a fairly significant project
                # that has to touch all the scrapli libraries
                self.session_lock.acquire()
        raise ScrapliTimeout(self.message)

    def _signal_raise_exception(self, signum: Any, frame: Any) -> None:
        """
        Signal method exception handler

        Args:
            signum: singum from the singal handler, unused here
            frame: frame from the signal handler, unused here

        Returns:
            N/A  # noqa: DAR202

        Raises:
            N/A

        """
        _, _ = signum, frame
        self._handle_timeout()

    async def asyncio_timeout(
        self, wrapped_func: Callable[..., Any], args: Any, kwargs: Any
    ) -> Any:
        """
        Asyncio method for timeouts

        Args:
            wrapped_func: function being decorated
            args: function being decorated args
            kwargs: function being decorated kwargs

        Returns:
            Any: result of decorated function

        Raises:
            N/A

        """
        try:
            return await asyncio.wait_for(
                wrapped_func(*args, **kwargs), timeout=self.timeout_duration
            )
        except asyncio.TimeoutError:
            self._handle_timeout()

    def signal_timeout(self, wrapped_func: Callable[..., Any], args: Any, kwargs: Any) -> Any:
        """
        Signal method for timeouts; does not work with system transport, on windows, or in threads

        Perhaps wondering why, if this doesnt work in so many places, do we have it? Great question!
        Because it is way way way way faster/less cpu intensive than the multiprocessing method!

        Args:
            wrapped_func: function being decorated
            args: function being decorated args
            kwargs: function being decorated kwargs

        Returns:
            Any: result of decorated function

        Raises:
            N/A

        """
        old = signal.signal(signal.SIGALRM, self._signal_raise_exception)
        signal.setitimer(signal.ITIMER_REAL, self.timeout_duration)
        try:
            return wrapped_func(*args, **kwargs)
        finally:
            if self.timeout_duration:
                signal.setitimer(signal.ITIMER_REAL, 0)
                signal.signal(signal.SIGALRM, old)

    def multiprocessing_timeout(
        self, wrapped_func: Callable[..., Any], args: Any, kwargs: Any
    ) -> Any:
        """
        Multiprocessing method for timeouts; works in threads and on windows

        Args:
            wrapped_func: function being decorated
            args: function being decorated args
            kwargs: function being decorated kwargs

        Returns:
            Any: result of decorated function

        Raises:
            N/A

        """
        with multiprocessing.pool.ThreadPool(processes=1) as pool:
            future = pool.apply_async(wrapped_func, args, kwargs)
            try:
                result = future.get(timeout=self.timeout_duration)
            except multiprocessing.context.TimeoutError:
                self._handle_timeout()
        return result


def requires_open_session() -> Callable[..., Any]:
    """
    Decorate an "operation" to require that the underlying scrapli session has been opened

    Args:
        N/A

    Returns:
        decorate: wrapped function

    Raises:
        ConnectionNotOpened: if scrapli connection has not been opened yet

    """

    def decorate(wrapped_func: Callable[..., Any]) -> Callable[..., Any]:
        def requires_open_session_wrapper(
            *args: Union[str, int],
            **kwargs: Dict[str, Union[str, int]],
        ) -> Any:
            try:
                return wrapped_func(*args, **kwargs)
            except AttributeError as exc:
                raise ConnectionNotOpened(
                    "Attempting to call method that requires an open connection, but connection is "
                    "not open. Call the `.open()` method of your connection object, or use a "
                    "context manager to ensue your connection has been opened."
                ) from exc

        return requires_open_session_wrapper

    return decorate


class TimeoutModifier:
    def __call__(self, wrapped_func: Callable[..., Any]) -> Callable[..., Any]:
        """
        Decorate an "operation" to modify the timeout_ops value for duration of that operation

        This decorator wraps send command/config ops and is used to allow users to set a
        `timeout_ops` value for the duration of a single method call -- this makes it so users don't
        need to manually set/reset the value

        Args:
            wrapped_func: function being decorated

        Returns:
            decorate: decorated func

        Raises:
            N/A

        """
        if asyncio.iscoroutinefunction(wrapped_func):

            async def decorate(*args: Any, **kwargs: Any) -> Any:
                scrapli_obj: Union["AsyncGenericDriver", "GenericDriver"] = args[0]
                if kwargs.get("timeout_ops", None) is None:
                    result = await wrapped_func(*args, **kwargs)
                elif kwargs.get("timeout_ops", scrapli_obj.timeout_ops) == scrapli_obj.timeout_ops:
                    result = await wrapped_func(*args, **kwargs)
                else:
                    base_timeout_ops = scrapli_obj.timeout_ops
                    scrapli_obj.timeout_ops = kwargs["timeout_ops"]
                    result = await wrapped_func(*args, **kwargs)
                    scrapli_obj.timeout_ops = base_timeout_ops
                return result

        else:

            def decorate(*args: Any, **kwargs: Any) -> Any:  # type: ignore
                scrapli_obj: Union["AsyncGenericDriver", "GenericDriver"] = args[0]
                if kwargs.get("timeout_ops", None) is None:
                    result = wrapped_func(*args, **kwargs)
                elif kwargs.get("timeout_ops", scrapli_obj.timeout_ops) == scrapli_obj.timeout_ops:
                    result = wrapped_func(*args, **kwargs)
                else:
                    base_timeout_ops = scrapli_obj.timeout_ops
                    scrapli_obj.timeout_ops = kwargs["timeout_ops"]
                    result = wrapped_func(*args, **kwargs)
                    scrapli_obj.timeout_ops = base_timeout_ops
                return result

        return decorate



import pytest

from scrapli.driver.core.arista_eos.async_driver import AsyncEOSDriver, eos_on_close, eos_on_open
from scrapli.driver.core.cisco_iosxe.async_driver import (
    AsyncIOSXEDriver,
    iosxe_on_close,
    iosxe_on_open,
)
from scrapli.driver.core.cisco_iosxr.async_driver import (
    AsyncIOSXRDriver,
    iosxr_on_close,
    iosxr_on_open,
)
from scrapli.driver.core.cisco_nxos.async_driver import AsyncNXOSDriver, nxos_on_close, nxos_on_open
from scrapli.driver.core.juniper_junos.async_driver import (
    AsyncJunosDriver,
    junos_on_close,
    junos_on_open,
)


def custom_open_close_func():
    return


@pytest.mark.parametrize(
    "attr_setup",
    [
        {"auth_secondary": "password", "on_open": None, "on_close": None},
        {
            "auth_secondary": "password",
            "on_open": custom_open_close_func,
            "on_close": custom_open_close_func,
        },
    ],
    ids=["default", "custom_open_close"],
)
@pytest.mark.parametrize(
    "platform",
    [
        (AsyncIOSXEDriver, iosxe_on_close, iosxe_on_open),
        (AsyncNXOSDriver, nxos_on_close, nxos_on_open),
        (AsyncIOSXRDriver, iosxr_on_close, iosxr_on_open),
        (AsyncEOSDriver, eos_on_close, eos_on_open),
        (AsyncJunosDriver, junos_on_close, junos_on_open),
    ],
    ids=["iosxe", "nxos", "iosxr", "eos", "junos"],
)
def test_core_driver_init(platform, attr_setup):
    driver_args = attr_setup
    driver = platform[0]
    on_close = driver_args["on_close"] or platform[1]
    on_open = driver_args["on_open"] or platform[2]
    conn = driver(host="myhost", transport="asyncssh", **driver_args)
    assert conn.auth_secondary == driver_args["auth_secondary"]
    assert conn.on_open == on_open
    assert conn.on_close == on_close



"""scrapli.driver.core.cisco_iosxr.base_driver"""
from scrapli.driver.base_network_driver import PrivilegeLevel

PRIVS = {
    "privilege_exec": (
        PrivilegeLevel(
            pattern=r"^[a-z0-9.\-_@/:]{1,63}#\s?$",
            name="privilege_exec",
            previous_priv="",
            deescalate="",
            escalate="",
            escalate_auth=False,
            escalate_prompt="",
        )
    ),
    "configuration": (
        PrivilegeLevel(
            pattern=r"^[a-z0-9.\-_@/:]{1,63}\(config[a-z0-9.\-@/:]{0,32}\)#\s?$",
            name="configuration",
            previous_priv="privilege_exec",
            deescalate="end",
            escalate="configure terminal",
            escalate_auth=False,
            escalate_prompt="",
        )
    ),
    "configuration_exclusive": (
        PrivilegeLevel(
            pattern=r"^[a-z0-9.\-_@/:]{1,63}\(config[a-z0-9.\-@/:]{0,32}\)#\s?$",
            name="configuration_exclusive",
            previous_priv="privilege_exec",
            deescalate="end",
            escalate="configure exclusive",
            escalate_auth=False,
            escalate_prompt="",
        )
    ),
}



import re

import pytest

from scrapli.driver.core.cisco_iosxe.driver import PRIVS


@pytest.mark.parametrize(
    "priv_pattern",
    [
        ("exec", "csr1000v>"),
        ("privilege_exec", "csr1000v#"),
        ("configuration", "csr1000v(config)#"),
        ("configuration", "csr1000v(conf-ssh-pubkey-data)#"),
        ("privilege_exec", "csr_1000v#"),
        ("configuration", "csr1000v(config-sg-tacacs+)#"),
    ],
    ids=[
        "base_prompt_exec",
        "base_prompt_privilege_exec",
        "base_prompt_configuration",
        "ssh_key_prompt",
        "underscore_privilege_exec",
        "tacacs_configuration",
    ],
)
def test_prompt_patterns(priv_pattern):
    priv_level_name = priv_pattern[0]
    prompt = priv_pattern[1]
    prompt_pattern = PRIVS.get(priv_level_name).pattern
    match = re.search(pattern=prompt_pattern, string=prompt, flags=re.M | re.I)
    assert match


@pytest.mark.parametrize(
    "prompt",
    [
        "password:",
        "Password:",
        "Enable password:",
    ],
    ids=["password:", "Password:", "Enable password:"],
)
def test_privilege_exec_escalation_prompt_patterns(prompt):
    escalation_pattern = PRIVS.get("privilege_exec").escalate_prompt
    match = re.search(pattern=escalation_pattern, string=prompt, flags=re.M | re.I)
    assert match



"""scrapli.transport.transport"""
import atexit
import time
from abc import ABC, abstractmethod
from concurrent.futures import ThreadPoolExecutor
from datetime import datetime

from scrapli.exceptions import ScrapliKeepaliveFailure
from scrapli.transport.base_transport import TransportBase


class Transport(TransportBase, ABC):
    @abstractmethod
    def open(self) -> None:
        """
        Open channel, acquire pty, request interactive shell

        Args:
            N/A

        Returns:
            N/A  # noqa: DAR202

        Raises:
            N/A

        """

    @abstractmethod
    def read(self) -> bytes:
        """
        Read data from the channel

        Args:
            N/A

        Returns:
            N/A  # noqa: DAR202

        Raises:
            N/A

        """

    @abstractmethod
    def write(self, channel_input: str) -> None:
        """
        Write data to the channel

        Args:
            channel_input: string to send to channel

        Returns:
            N/A  # noqa: DAR202

        Raises:
            N/A

        """

    def _session_keepalive(self) -> None:
        """
        Spawn keepalive thread for transport session

        Args:
            N/A

        Returns:
            N/A  # noqa: DAR202

        Raises:
            N/A

        """
        if not self.keepalive:
            return
        pool = ThreadPoolExecutor()
        if self.keepalive_type == "network":
            pool.submit(self._keepalive_network)
        else:
            pool.submit(self._keepalive_standard)

    def _keepalive_network(self) -> None:
        """
        Send "in band" keepalives to devices.

        Generally used with "network" devices which do not have native keepalive support. This will
        try to acquire a session lock and send an innocuous character -- such as CTRL+E -- to keep
        the device "exec-timeout" (in network-y words) from expiring.

        Args:
            N/A

        Returns:
            N/A  # noqa: DAR202

        Raises:
            ScrapliKeepaliveFailure: if scrapli cant unlock and send keepalive in less than 3 *
                the keepalive_interval

        """

        def kill_transport() -> None:
            if self.isalive():
                self.close()

        # ensure the transport is closed so keepalive can terminate if main thread tries to exit
        atexit.register(kill_transport)

        lock_counter = 0
        last_keepalive = datetime.now()
        while True:
            if not self.isalive():
                return
            diff = datetime.now() - last_keepalive
            if diff.seconds >= self.keepalive_interval:
                if not self.session_lock.locked():
                    self.logger.debug(
                        f"Sending 'network' keepalive with pattern {repr(self.keepalive_pattern)}."
                    )
                    lock_counter = 0
                    self.session_lock.acquire()
                    self.write(channel_input=self.keepalive_pattern)
                    self.session_lock.release()
                    last_keepalive = datetime.now()
                else:
                    lock_counter += 1
                    if lock_counter >= 3:
                        self.logger.info(
                            f"Keepalive thread missed {lock_counter} consecutive keepalives."
                        )
            if diff.seconds > self.keepalive_interval * 3:
                msg = (
                    "Keepalive thread has failed to send a keepalive in greater than three "
                    "times the keepalive interval!"
                )
                self.logger.critical(msg)
                raise ScrapliKeepaliveFailure(msg)
            time.sleep(self.keepalive_interval / 10)

    @abstractmethod
    def _keepalive_standard(self) -> None:
        """
        Send "out of band" (protocol level) keepalives to devices.

        Args:
            N/A

        Returns:
            N/A  # noqa: DAR202

        Raises:
            N/A

        """



import time

import pytest

from scrapli.exceptions import ScrapliAuthenticationFailed

from .test_data.devices import DEVICES, INVALID_PRIVATE_KEY, PRIVATE_KEY
from .test_data.test_cases import TEST_CASES


class TestDevice:
    def test_get_prompt(self, nix_conn, transport):
        prompt = nix_conn.channel.get_prompt()
        assert prompt == TEST_CASES["linux"]["get_prompt"]

    @pytest.mark.parametrize(
        "strip_prompt",
        [True, False],
        ids=["strip_prompt", "no_strip_prompt"],
    )
    def test_send_command_short(self, nix_conn, transport, strip_prompt):
        command = TEST_CASES["linux"]["send_command_short"]["command"]
        expected_type = "expected_no_strip" if not strip_prompt else "expected_strip"
        expected_response = TEST_CASES["linux"]["send_command_short"][expected_type]
        _, response = nix_conn.channel.send_input(channel_input=command, strip_prompt=strip_prompt)
        assert response.decode() == expected_response

    @pytest.mark.parametrize(
        "strip_prompt",
        [True, False],
        ids=["strip_prompt", "no_strip_prompt"],
    )
    def test_send_command_long(self, nix_conn, transport, strip_prompt):
        command = TEST_CASES["linux"]["send_command_long"]["command"]
        expected_type = "expected_no_strip" if not strip_prompt else "expected_strip"
        expected_response = TEST_CASES["linux"]["send_command_long"][expected_type]
        _, response = nix_conn.channel.send_input(channel_input=command, strip_prompt=strip_prompt)
        assert response.decode() == expected_response

    def test_isalive_and_close(self, nix_conn, transport):
        assert nix_conn.isalive() is True
        nix_conn.close()
        # unsure why but w/out a tiny sleep pytest just plows ahead and the connection doesnt
        # close in time for the next assert
        time.sleep(0.1)
        assert nix_conn.isalive() is False


class TestGenericDevice:
    def test_get_prompt(self, nix_conn_generic, transport):
        prompt = nix_conn_generic.channel.get_prompt()
        assert prompt == TEST_CASES["linux"]["get_prompt"]

    @pytest.mark.parametrize(
        "strip_prompt",
        [True, False],
        ids=["strip_prompt", "no_strip_prompt"],
    )
    def test_send_command_short(self, nix_conn_generic, transport, strip_prompt):
        command = TEST_CASES["linux"]["send_command_short"]["command"]
        expected_type = "expected_no_strip" if not strip_prompt else "expected_strip"
        expected_response = TEST_CASES["linux"]["send_command_short"][expected_type]
        response = nix_conn_generic.send_command(command=command, strip_prompt=strip_prompt)
        assert response.result == expected_response

    @pytest.mark.parametrize(
        "strip_prompt",
        [True, False],
        ids=["strip_prompt", "no_strip_prompt"],
    )
    def test_send_command_long(self, nix_conn_generic, transport, strip_prompt):
        command = TEST_CASES["linux"]["send_command_long"]["command"]
        expected_type = "expected_no_strip" if not strip_prompt else "expected_strip"
        expected_response = TEST_CASES["linux"]["send_command_long"][expected_type]
        response = nix_conn_generic.send_command(command=command, strip_prompt=strip_prompt)
        assert response.result == expected_response

    @pytest.mark.parametrize(
        "strip_prompt",
        [True, False],
        ids=["strip_prompt", "no_strip_prompt"],
    )
    def test_send_commands(self, nix_conn_generic, transport, strip_prompt):
        commands = []
        expected_responses = []
        commands.append(TEST_CASES["linux"]["send_command_short"]["command"])
        commands.append(TEST_CASES["linux"]["send_command_long"]["command"])
        expected_type = "expected_no_strip" if not strip_prompt else "expected_strip"
        expected_responses.append(TEST_CASES["linux"]["send_command_short"][expected_type])
        expected_responses.append(TEST_CASES["linux"]["send_command_long"][expected_type])
        responses = nix_conn_generic.send_commands(commands=commands, strip_prompt=strip_prompt)
        for expected_response, response in zip(expected_responses, responses):
            assert response.result == expected_response

    def test_isalive_and_close(self, nix_conn_generic, transport):
        assert nix_conn_generic.isalive() is True
        nix_conn_generic.close()
        # unsure why but w/out a tiny sleep pytest just plows ahead and the connection doesnt
        # close in time for the next assert
        time.sleep(0.1)
        assert nix_conn_generic.isalive() is False


class TestNetworkDevice:
    @pytest.mark.parametrize(
        "priv_level",
        ["exec", "privilege_exec", "configuration"],
        ids=["exec", "privilege_exec", "configuration"],
    )
    def test_get_prompt_and_acquire_priv(self, conn, priv_level, device_type, transport):
        if TEST_CASES[device_type]["get_prompt"][priv_level] is None:
            pytest.skip(f"Priv level {priv_level} for device type {device_type} not tested")
        conn.acquire_priv(priv_level)
        prompt = conn.get_prompt()
        assert prompt == TEST_CASES[device_type]["get_prompt"][priv_level]

    @pytest.mark.parametrize(
        "strip_prompt",
        [True, False],
        ids=["strip_prompt", "no_strip_prompt"],
    )
    def test_send_command_short(self, conn, device_type, transport, strip_prompt):
        command = TEST_CASES[device_type]["send_command_short"]["command"]
        expected_type = "expected_no_strip" if not strip_prompt else "expected_strip"
        expected_response = TEST_CASES[device_type]["send_command_short"][expected_type]
        sanitize_response = TEST_CASES[device_type]["sanitize_response"]
        response = conn.send_command(command=command, strip_prompt=strip_prompt)
        assert sanitize_response(response.result) == expected_response

    @pytest.mark.parametrize(
        "strip_prompt",
        [True, False],
        ids=["strip_prompt", "no_strip_prompt"],
    )
    def test_send_command_long(self, conn, device_type, transport, strip_prompt):
        command = TEST_CASES[device_type]["send_command_long"]["command"]
        expected_type = "expected_no_strip" if not strip_prompt else "expected_strip"
        expected_response = TEST_CASES[device_type]["send_command_long"][expected_type]
        sanitize_response = TEST_CASES[device_type]["sanitize_response"]
        response = conn.send_command(command=command, strip_prompt=strip_prompt)
        assert sanitize_response(response.result) == expected_response

    @pytest.mark.parametrize(
        "strip_prompt",
        [True, False],
        ids=["strip_prompt", "no_strip_prompt"],
    )
    def test_send_commands(self, conn, device_type, transport, strip_prompt):
        commands = []
        expected_responses = []
        commands.append(TEST_CASES[device_type]["send_command_short"]["command"])
        commands.append(TEST_CASES[device_type]["send_command_long"]["command"])
        expected_type = "expected_no_strip" if not strip_prompt else "expected_strip"
        expected_responses.append(TEST_CASES[device_type]["send_command_short"][expected_type])
        expected_responses.append(TEST_CASES[device_type]["send_command_long"][expected_type])
        sanitize_response = TEST_CASES[device_type]["sanitize_response"]
        responses = conn.send_commands(commands=commands, strip_prompt=strip_prompt)
        for expected_response, response in zip(expected_responses, responses):
            assert sanitize_response(response.result) == expected_response

    @pytest.mark.parametrize(
        "strip_prompt",
        [True, False],
        ids=["strip_prompt", "no_strip_prompt"],
    )
    def test_send_commands_from_file(self, conn, device_type, transport, strip_prompt):
        file = TEST_CASES[device_type]["send_commands_from_file"]["file"]
        expected_type = "expected_no_strip" if not strip_prompt else "expected_strip"
        expected_responses = TEST_CASES[device_type]["send_commands_from_file"][expected_type]
        sanitize_response = TEST_CASES[device_type]["sanitize_response"]
        responses = conn.send_commands_from_file(file=file, strip_prompt=strip_prompt)
        for expected_response, response in zip(expected_responses, responses):
            assert sanitize_response(response.result) == expected_response

    def test_send_commands_stop_on_failed(self, conn, device_type, transport):
        commands = TEST_CASES[device_type]["send_commands_error"]["commands"]
        responses = conn.send_commands(commands=commands, stop_on_failed=True)
        assert len(responses) == 2
        assert responses[0].failed is False
        assert responses[1].failed is True

    def test_send_interactive_normal_response(self, conn, device_type, transport):
        if TEST_CASES[device_type]["send_interactive_normal_response"] is None:
            pytest.skip(
                f"send_interactive_normal_response for device type {device_type} not tested"
            )
        interact = TEST_CASES[device_type]["send_interactive_normal_response"]["command"]
        expected_response = TEST_CASES[device_type]["send_interactive_normal_response"]["expected"]
        response = conn.send_interactive(interact_events=interact)
        assert response.result == expected_response

    def test_send_interactive_hidden_response(self, conn, device_type, transport):
        if TEST_CASES[device_type]["send_interactive_hidden_response"] is None:
            pytest.skip(
                f"send_interactive_hidden_response for device type {device_type} not tested"
            )
        interact = TEST_CASES[device_type]["send_interactive_hidden_response"]["command"]
        expected_response = TEST_CASES[device_type]["send_interactive_hidden_response"]["expected"]
        response = conn.send_interactive(interact_events=interact)
        assert response.result == expected_response

    @pytest.mark.parametrize(
        "strip_prompt",
        [True, False],
        ids=["strip_prompt", "no_strip_prompt"],
    )
    def test_send_config(self, conn, device_type, transport, strip_prompt):
        config = TEST_CASES[device_type]["send_config"]["configs"]
        expected_type = "expected_no_strip" if not strip_prompt else "expected_strip"
        expected_response = TEST_CASES[device_type]["send_config"][expected_type]
        verification = TEST_CASES[device_type]["send_config"]["verification"]
        expected_verification = TEST_CASES[device_type]["send_config"][
            f"verification_{expected_type}"
        ]
        teardown_configs = TEST_CASES[device_type]["send_config"]["teardown_configs"]
        sanitize_response = TEST_CASES[device_type]["sanitize_response"]
        response = conn.send_config(config=config, strip_prompt=strip_prompt)
        assert sanitize_response(response.result) == expected_response
        verification_response = conn.send_command(command=verification, strip_prompt=strip_prompt)
        assert sanitize_response(verification_response.result) == expected_verification
        if isinstance(teardown_configs, list):
            conn.send_configs(configs=teardown_configs)
        else:
            conn.send_config(config=teardown_configs)

    @pytest.mark.parametrize(
        "strip_prompt",
        [True, False],
        ids=["strip_prompt", "no_strip_prompt"],
    )
    def test_send_configs(self, conn, device_type, transport, strip_prompt):
        configs = TEST_CASES[device_type]["send_configs"]["configs"]
        expected_type = "expected_no_strip" if not strip_prompt else "expected_strip"
        expected_responses = TEST_CASES[device_type]["send_configs"][expected_type]
        verification = TEST_CASES[device_type]["send_configs"]["verification"]
        expected_verification = TEST_CASES[device_type]["send_configs"][
            f"verification_{expected_type}"
        ]
        teardown_configs = TEST_CASES[device_type]["send_configs"]["teardown_configs"]
        sanitize_response = TEST_CASES[device_type]["sanitize_response"]
        responses = conn.send_configs(configs=configs, strip_prompt=strip_prompt)
        for expected_response, response in zip(expected_responses, responses):
            assert response.result == expected_response
        verification_response = conn.send_command(command=verification, strip_prompt=strip_prompt)
        assert sanitize_response(verification_response.result) == expected_verification
        if isinstance(teardown_configs, list):
            conn.send_configs(configs=teardown_configs)
        else:
            conn.send_config(config=teardown_configs)

    @pytest.mark.parametrize(
        "strip_prompt",
        [True, False],
        ids=["strip_prompt", "no_strip_prompt"],
    )
    def test_send_configs_from_file(self, conn, device_type, transport, strip_prompt):
        file = TEST_CASES[device_type]["send_configs_from_file"]["file"]
        expected_type = "expected_no_strip" if not strip_prompt else "expected_strip"
        expected_responses = TEST_CASES[device_type]["send_configs"][expected_type]
        verification = TEST_CASES[device_type]["send_configs"]["verification"]
        expected_verification = TEST_CASES[device_type]["send_configs"][
            f"verification_{expected_type}"
        ]
        teardown_configs = TEST_CASES[device_type]["send_configs"]["teardown_configs"]
        sanitize_response = TEST_CASES[device_type]["sanitize_response"]
        responses = conn.send_configs_from_file(file=file, strip_prompt=strip_prompt)
        for expected_response, response in zip(expected_responses, responses):
            assert response.result == expected_response
        verification_response = conn.send_command(command=verification, strip_prompt=strip_prompt)
        assert sanitize_response(verification_response.result) == expected_verification
        if isinstance(teardown_configs, list):
            conn.send_configs(configs=teardown_configs)
        else:
            conn.send_config(config=teardown_configs)

    def test_send_configs_stop_on_failed(self, conn, device_type, transport):
        if TEST_CASES[device_type]["send_configs_error"] is None:
            pytest.skip(
                f"test_send_configs_stop_on_failed for device type {device_type} not tested yet, need to overhaul privilege levels..."
            )
        configs = TEST_CASES[device_type]["send_configs_error"]["configs"]
        teardown_configs = TEST_CASES[device_type]["send_configs_error"]["teardown_configs"]
        responses = conn.send_configs(configs=configs, stop_on_failed=True)
        assert len(responses) == 2
        assert responses[0].failed is False
        assert responses[1].failed is True
        if isinstance(teardown_configs, list):
            conn.send_configs(configs=teardown_configs)
        else:
            conn.send_config(config=teardown_configs)

    def test_isalive_and_close(self, conn, device_type, transport):
        assert conn.isalive() is True
        conn.close()
        # unsure why but w/out a tiny sleep pytest just plows ahead and the connection doesnt
        # close in time for the next assert
        time.sleep(0.5)
        assert conn.isalive() is False


@pytest.mark.asyncio
class TestAsyncNetworkDevice:
    @pytest.mark.parametrize(
        "priv_level",
        ["exec", "privilege_exec", "configuration"],
        ids=["exec", "privilege_exec", "configuration"],
    )
    async def test_get_prompt_and_acquire_priv(self, async_conn, priv_level, device_type):
        if TEST_CASES[device_type]["get_prompt"][priv_level] is None:
            pytest.skip(f"Priv level {priv_level} for device type {device_type} not tested")
        await async_conn.acquire_priv(priv_level)
        prompt = await async_conn.get_prompt()
        assert prompt == TEST_CASES[device_type]["get_prompt"][priv_level]

    @pytest.mark.parametrize(
        "strip_prompt",
        [True, False],
        ids=["strip_prompt", "no_strip_prompt"],
    )
    async def test_send_command_short(self, async_conn, device_type, strip_prompt):
        command = TEST_CASES[device_type]["send_command_short"]["command"]
        expected_type = "expected_no_strip" if not strip_prompt else "expected_strip"
        expected_response = TEST_CASES[device_type]["send_command_short"][expected_type]
        sanitize_response = TEST_CASES[device_type]["sanitize_response"]
        response = await async_conn.send_command(command=command, strip_prompt=strip_prompt)
        assert sanitize_response(response.result) == expected_response

    @pytest.mark.parametrize(
        "strip_prompt",
        [True, False],
        ids=["strip_prompt", "no_strip_prompt"],
    )
    async def test_send_command_long(self, async_conn, device_type, strip_prompt):
        command = TEST_CASES[device_type]["send_command_long"]["command"]
        expected_type = "expected_no_strip" if not strip_prompt else "expected_strip"
        expected_response = TEST_CASES[device_type]["send_command_long"][expected_type]
        sanitize_response = TEST_CASES[device_type]["sanitize_response"]
        response = await async_conn.send_command(command=command, strip_prompt=strip_prompt)
        assert sanitize_response(response.result) == expected_response

    @pytest.mark.parametrize(
        "strip_prompt",
        [True, False],
        ids=["strip_prompt", "no_strip_prompt"],
    )
    async def test_send_commands(self, async_conn, device_type, strip_prompt):
        commands = []
        expected_responses = []
        commands.append(TEST_CASES[device_type]["send_command_short"]["command"])
        commands.append(TEST_CASES[device_type]["send_command_long"]["command"])
        expected_type = "expected_no_strip" if not strip_prompt else "expected_strip"
        expected_responses.append(TEST_CASES[device_type]["send_command_short"][expected_type])
        expected_responses.append(TEST_CASES[device_type]["send_command_long"][expected_type])
        sanitize_response = TEST_CASES[device_type]["sanitize_response"]
        responses = await async_conn.send_commands(commands=commands, strip_prompt=strip_prompt)
        for expected_response, response in zip(expected_responses, responses):
            assert sanitize_response(response.result) == expected_response

    @pytest.mark.parametrize(
        "strip_prompt",
        [True, False],
        ids=["strip_prompt", "no_strip_prompt"],
    )
    async def test_send_commands_from_file(self, async_conn, device_type, strip_prompt):
        file = TEST_CASES[device_type]["send_commands_from_file"]["file"]
        expected_type = "expected_no_strip" if not strip_prompt else "expected_strip"
        expected_responses = TEST_CASES[device_type]["send_commands_from_file"][expected_type]
        sanitize_response = TEST_CASES[device_type]["sanitize_response"]
        responses = await async_conn.send_commands_from_file(file=file, strip_prompt=strip_prompt)
        for expected_response, response in zip(expected_responses, responses):
            assert sanitize_response(response.result) == expected_response

    async def test_send_commands_stop_on_failed(self, async_conn, device_type):
        commands = TEST_CASES[device_type]["send_commands_error"]["commands"]
        responses = await async_conn.send_commands(commands=commands, stop_on_failed=True)
        assert len(responses) == 2
        assert responses[0].failed is False
        assert responses[1].failed is True

    async def test_send_interactive_normal_response(self, async_conn, device_type):
        if TEST_CASES[device_type]["send_interactive_normal_response"] is None:
            pytest.skip(
                f"send_interactive_normal_response for device type {device_type} not tested"
            )
        interact = TEST_CASES[device_type]["send_interactive_normal_response"]["command"]
        expected_response = TEST_CASES[device_type]["send_interactive_normal_response"]["expected"]
        response = await async_conn.send_interactive(interact_events=interact)
        assert response.result == expected_response

    async def test_send_interactive_hidden_response(self, async_conn, device_type):
        if TEST_CASES[device_type]["send_interactive_hidden_response"] is None:
            pytest.skip(
                f"send_interactive_hidden_response for device type {device_type} not tested"
            )
        interact = TEST_CASES[device_type]["send_interactive_hidden_response"]["command"]
        expected_response = TEST_CASES[device_type]["send_interactive_hidden_response"]["expected"]
        response = await async_conn.send_interactive(interact_events=interact)
        assert response.result == expected_response

    @pytest.mark.parametrize(
        "strip_prompt",
        [True, False],
        ids=["strip_prompt", "no_strip_prompt"],
    )
    async def test_send_config(self, async_conn, device_type, strip_prompt):
        config = TEST_CASES[device_type]["send_config"]["configs"]
        expected_type = "expected_no_strip" if not strip_prompt else "expected_strip"
        expected_response = TEST_CASES[device_type]["send_config"][expected_type]
        verification = TEST_CASES[device_type]["send_config"]["verification"]
        expected_verification = TEST_CASES[device_type]["send_config"][
            f"verification_{expected_type}"
        ]
        teardown_configs = TEST_CASES[device_type]["send_config"]["teardown_configs"]
        sanitize_response = TEST_CASES[device_type]["sanitize_response"]
        response = await async_conn.send_config(config=config, strip_prompt=strip_prompt)
        assert sanitize_response(response.result) == expected_response
        verification_response = await async_conn.send_command(
            command=verification, strip_prompt=strip_prompt
        )
        assert sanitize_response(verification_response.result) == expected_verification
        if isinstance(teardown_configs, list):
            await async_conn.send_configs(configs=teardown_configs)
        else:
            await async_conn.send_config(config=teardown_configs)

    @pytest.mark.parametrize(
        "strip_prompt",
        [True, False],
        ids=["strip_prompt", "no_strip_prompt"],
    )
    async def test_send_configs(self, async_conn, device_type, strip_prompt):
        configs = TEST_CASES[device_type]["send_configs"]["configs"]
        expected_type = "expected_no_strip" if not strip_prompt else "expected_strip"
        expected_responses = TEST_CASES[device_type]["send_configs"][expected_type]
        verification = TEST_CASES[device_type]["send_configs"]["verification"]
        expected_verification = TEST_CASES[device_type]["send_configs"][
            f"verification_{expected_type}"
        ]
        teardown_configs = TEST_CASES[device_type]["send_configs"]["teardown_configs"]
        sanitize_response = TEST_CASES[device_type]["sanitize_response"]
        responses = await async_conn.send_configs(configs=configs, strip_prompt=strip_prompt)
        for expected_response, response in zip(expected_responses, responses):
            assert response.result == expected_response
        verification_response = await async_conn.send_command(
            command=verification, strip_prompt=strip_prompt
        )
        assert sanitize_response(verification_response.result) == expected_verification
        if isinstance(teardown_configs, list):
            await async_conn.send_configs(configs=teardown_configs)
        else:
            await async_conn.send_config(config=teardown_configs)

    @pytest.mark.parametrize(
        "strip_prompt",
        [True, False],
        ids=["strip_prompt", "no_strip_prompt"],
    )
    async def test_send_configs_from_file(self, async_conn, device_type, strip_prompt):
        file = TEST_CASES[device_type]["send_configs_from_file"]["file"]
        expected_type = "expected_no_strip" if not strip_prompt else "expected_strip"
        expected_responses = TEST_CASES[device_type]["send_configs"][expected_type]
        verification = TEST_CASES[device_type]["send_configs"]["verification"]
        expected_verification = TEST_CASES[device_type]["send_configs"][
            f"verification_{expected_type}"
        ]
        teardown_configs = TEST_CASES[device_type]["send_configs"]["teardown_configs"]
        sanitize_response = TEST_CASES[device_type]["sanitize_response"]
        responses = await async_conn.send_configs_from_file(file=file, strip_prompt=strip_prompt)
        for expected_response, response in zip(expected_responses, responses):
            assert response.result == expected_response
        verification_response = await async_conn.send_command(
            command=verification, strip_prompt=strip_prompt
        )
        assert sanitize_response(verification_response.result) == expected_verification
        if isinstance(teardown_configs, list):
            await async_conn.send_configs(configs=teardown_configs)
        else:
            await async_conn.send_config(config=teardown_configs)

    async def test_send_configs_stop_on_failed(self, async_conn, device_type):
        if TEST_CASES[device_type]["send_configs_error"] is None:
            pytest.skip(
                f"test_send_configs_stop_on_failed for device type {device_type} not tested yet, need to overhaul privilege levels..."
            )
        configs = TEST_CASES[device_type]["send_configs_error"]["configs"]
        teardown_configs = TEST_CASES[device_type]["send_configs_error"]["teardown_configs"]
        responses = await async_conn.send_configs(configs=configs, stop_on_failed=True)
        assert len(responses) == 2
        assert responses[0].failed is False
        assert responses[1].failed is True
        if isinstance(teardown_configs, list):
            await async_conn.send_configs(configs=teardown_configs)
        else:
            await async_conn.send_config(config=teardown_configs)

    async def test_isalive_and_close(self, async_conn, device_type):
        assert async_conn.isalive() is True
        await async_conn.close()
        # unsure why but w/out a tiny sleep pytest just plows ahead and the connection doesnt
        # close in time for the next assert
        time.sleep(0.5)
        assert async_conn.isalive() is False


def test_context_manager(device_type, transport):
    if device_type == "arista_eos" and transport == "ssh2":
        pytest.skip(
            "SSH2 (on pypi) doesn't support keyboard interactive auth, skipping ssh2 for arista_eos testing"
        )

    device = DEVICES[device_type].copy()
    driver = device.pop("driver")
    device.pop("base_config")
    device.pop("async_driver")

    port = 22
    if transport == "telnet":
        port = 23

    device["port"] = port
    device["transport"] = transport
    device["timeout_socket"] = 5
    device["timeout_transport"] = 5
    device["timeout_ops"] = 5

    with driver(**device) as conn:
        assert conn.isalive() is True
    # unsure why but w/out a tiny sleep pytest just plows ahead and the connection doesnt
    # close in time for the next assert
    time.sleep(0.5)
    assert conn.isalive() is False


def test_public_key_auth(device_type, transport):
    if device_type != "cisco_iosxe" or transport == "telnet":
        pytest.skip("public key auth only tested against iosxe at the moment, and never on telnet!")

    device = DEVICES[device_type].copy()
    driver = device.pop("driver")
    device.pop("base_config")
    device.pop("async_driver")

    device["transport"] = transport
    device["timeout_socket"] = 5
    device["timeout_transport"] = 5
    device["timeout_ops"] = 5
    device["auth_private_key"] = PRIVATE_KEY

    with driver(**device) as conn:
        assert conn.isalive() is True
    # unsure why but w/out a tiny sleep pytest just plows ahead and the connection doesnt
    # close in time for the next assert
    time.sleep(0.2)
    assert conn.isalive() is False


def test_public_key_auth_failure(device_type, transport):
    if device_type != "cisco_iosxe" or transport == "telnet":
        pytest.skip("public key auth only tested against iosxe at the moment, and never on telnet!")

    if device_type != "cisco_iosxe" or transport == "system":
        pytest.skip(
            "systemssh raises a different exception as it cant distinguish between auth types as well as the other transports!"
        )

    device = DEVICES[device_type].copy()
    driver = device.pop("driver")
    device.pop("base_config")
    device.pop("async_driver")

    device["transport"] = transport
    device["timeout_socket"] = 2
    device["timeout_transport"] = 2
    device["timeout_ops"] = 2
    device["auth_private_key"] = INVALID_PRIVATE_KEY
    device.pop("auth_password")
    conn = driver(**device)

    with pytest.raises(ScrapliAuthenticationFailed) as exc:
        conn.open()
    assert str(exc.value) == (
        f"Failed to authenticate to host 172.18.0.11 with private key `{INVALID_PRIVATE_KEY}`. "
        "Unable to continue authentication, missing username, password, or both."
    )


def test_public_key_auth_failure_systemssh(device_type, transport):
    if device_type != "cisco_iosxe" or transport != "system":
        pytest.skip(
            "systemssh raises a different exception as it cant distinguish between auth types as well as the other transports!"
        )

    device = DEVICES[device_type].copy()
    driver = device.pop("driver")
    device.pop("base_config")
    device.pop("async_driver")

    device["transport"] = transport
    device["timeout_socket"] = 2
    device["timeout_transport"] = 2
    device["timeout_ops"] = 5
    device["auth_private_key"] = INVALID_PRIVATE_KEY
    device.pop("auth_password")
    conn = driver(**device)

    with pytest.raises(ScrapliAuthenticationFailed) as exc:
        conn.open()
    assert str(exc.value) == "Authentication to host 172.18.0.11 failed"



"""examples.test_examples"""
import pytest

from .async_usage import async_iosxe_driver
from .basic_usage import generic_driver, iosxe_driver, scrapli_driver
from .configuration_modes import eos_configure_session, iosxr_configure_exclusive
from .logging import basic_logging
from .ssh_keys import ssh_keys
from .structured_data import structured_data_genie

# simply making sure the examples don't blow up!


@pytest.mark.asyncio
async def test_async_usage():
    await async_iosxe_driver.main()


@pytest.mark.parametrize(
    "example_script",
    [generic_driver, iosxe_driver, scrapli_driver],
    ids=["generic_driver", "iosxe_driver", "scrapli_driver"],
)
def test_basic_usage(example_script):
    example_script.main()


@pytest.mark.parametrize(
    "example_script",
    [iosxr_configure_exclusive, eos_configure_session],
    ids=["iosxr_configure_exclusive", "eos_configure_session"],
)
def test_configuration_modes(example_script):
    example_script.main()


def test_logging():
    basic_logging.main()


def test_ssh_keys():
    ssh_keys.main()


@pytest.mark.parametrize(
    "example_script",
    [structured_data_genie, structured_data_genie],
    ids=["genie", "textfsm"],
)
def test_structured_data(example_script):
    example_script.main()



"""scrapli.driver.core.cisco_nxos"""
from scrapli.driver.core.cisco_nxos.async_driver import AsyncNXOSDriver
from scrapli.driver.core.cisco_nxos.driver import NXOSDriver

__all__ = (
    "AsyncNXOSDriver",
    "NXOSDriver",
)



import os
import sys
from pathlib import Path

import pytest

import scrapli
from scrapli.ssh_config import Host, SSHConfig, SSHKnownHosts

TEST_DATA_DIR = f"{Path(scrapli.__file__).parents[1]}/tests/test_data"


def test_str():
    ssh_conf = SSHConfig(f"{TEST_DATA_DIR}/files/_ssh_config")
    assert str(ssh_conf) == "SSHConfig Object"


@pytest.mark.skipif(
    sys.platform.startswith("win"), reason="not currently testing ssh file resolution on windows"
)
def test_repr():
    ssh_conf = SSHConfig(f"{TEST_DATA_DIR}/files/_ssh_config")
    assert repr(ssh_conf) == (
        f"SSHConfig {{'ssh_config_file': '{TEST_DATA_DIR}/files/_ssh_config', 'hosts': {{'1.2.3.4 someswitch1': Host {{'hosts': '1.2.3.4 someswitch1', 'hostname': 'someswitch1.bogus.com', 'port': 1234, 'user': 'carl', 'address_family': None, 'bind_address': None, 'connect_timeout': None, 'identities_only': 'yes', 'identity_file': '{os.path.expanduser('~')}/.ssh/mysshkey', 'keyboard_interactive': None, 'password_authentication': None, 'preferred_authentication': None}}, 'someswitch?': Host {{'hosts': 'someswitch?', 'hostname': 'someswitch1.bogus.com', 'port': 1234, 'user': 'notcarl', 'address_family': None, 'bind_address': None, 'connect_timeout': None, 'identities_only': 'yes', 'identity_file': '{os.path.expanduser('~')}/.ssh/mysshkey', 'keyboard_interactive': None, 'password_authentication': None, 'preferred_authentication': None}}, 'scrapli': Host {{'hosts': 'scrapli', 'hostname': None, 'port': None, 'user': 'scrapli', 'address_family': None, 'bind_address': None, 'connect_timeout': None, 'identities_only': None, 'identity_file': '{os.path.expanduser('~')}/.ssh/lastresortkey', 'keyboard_interactive': None, 'password_authentication': None, 'preferred_authentication': None}}, '*': Host {{'hosts': '*', 'hostname': None, 'port': None, 'user': 'somebodyelse', 'address_family': None, 'bind_address': None, 'connect_timeout': None, 'identities_only': None, 'identity_file': '{os.path.expanduser('~')}/.ssh/lastresortkey', 'keyboard_interactive': None, 'password_authentication': None, 'preferred_authentication': None}}}}}}"
    )


def test_bool_true():
    ssh_conf = SSHConfig(f"{TEST_DATA_DIR}/files/_ssh_config")
    assert bool(ssh_conf) is True


def test_bool_false():
    ssh_conf = SSHConfig(f"{TEST_DATA_DIR}/files/__init__.py")
    assert bool(ssh_conf) is False


def test_host__str():
    host = Host()
    assert str(host) == "Host: "


@pytest.mark.skipif(
    sys.platform.startswith("win"), reason="not currently testing ssh file resolution on windows"
)
def test_host__repr():
    ssh_conf = SSHConfig(f"{TEST_DATA_DIR}/files/_ssh_config")
    assert repr(ssh_conf.hosts["1.2.3.4 someswitch1"]) == (
        "Host {'hosts': '1.2.3.4 someswitch1', 'hostname': 'someswitch1.bogus.com', 'port': "
        "1234, 'user': 'carl', 'address_family': None, 'bind_address': None, 'connect_timeout': "
        f"None, 'identities_only': 'yes', 'identity_file': '{os.path.expanduser('~')}/.ssh/mysshkey', "
        "'keyboard_interactive': None, 'password_authentication': None, 'preferred_authentication': "
        "None}"
    )


def test_init_ssh_config_file_explicit():
    ssh_conf = SSHConfig(f"{TEST_DATA_DIR}/files/_ssh_config")
    with open(f"{TEST_DATA_DIR}/files/_ssh_config", "r") as f:
        ssh_config_file = f.read()
    assert ssh_conf.ssh_config == ssh_config_file


def test_init_ssh_config_file_no_config_file(fs):
    ssh_conf = SSHConfig("")
    # should only have a single splat host w/ all values set to None/empty
    assert ["*"] == list(ssh_conf.hosts.keys())
    assert ssh_conf.hosts["*"].hosts == "*"
    assert ssh_conf.hosts["*"].hostname is None
    assert ssh_conf.hosts["*"].port is None
    assert ssh_conf.hosts["*"].user == ""
    assert ssh_conf.hosts["*"].address_family is None
    assert ssh_conf.hosts["*"].bind_address is None
    assert ssh_conf.hosts["*"].connect_timeout is None
    assert ssh_conf.hosts["*"].identities_only is None
    assert ssh_conf.hosts["*"].identity_file is None
    assert ssh_conf.hosts["*"].keyboard_interactive is None
    assert ssh_conf.hosts["*"].password_authentication is None
    assert ssh_conf.hosts["*"].preferred_authentication is None


def test_init_ssh_config_file_no_hosts():
    ssh_conf = SSHConfig(f"{TEST_DATA_DIR}/files/__init__.py")
    assert ["*"] == list(ssh_conf.hosts.keys())
    assert ssh_conf.hosts["*"].hosts == "*"
    assert ssh_conf.hosts["*"].hostname is None
    assert ssh_conf.hosts["*"].port is None
    assert ssh_conf.hosts["*"].user == ""
    assert ssh_conf.hosts["*"].address_family is None
    assert ssh_conf.hosts["*"].bind_address is None
    assert ssh_conf.hosts["*"].connect_timeout is None
    assert ssh_conf.hosts["*"].identities_only is None
    assert ssh_conf.hosts["*"].identity_file is None
    assert ssh_conf.hosts["*"].keyboard_interactive is None
    assert ssh_conf.hosts["*"].password_authentication is None
    assert ssh_conf.hosts["*"].preferred_authentication is None


@pytest.mark.skipif(
    sys.platform.startswith("win"), reason="not currently testing ssh file resolution on windows"
)
def test_host_lookup_exact_host():
    ssh_conf = SSHConfig(f"{TEST_DATA_DIR}/files/_ssh_config")
    host = ssh_conf.lookup("1.2.3.4 someswitch1")
    assert repr(host) == (
        "Host {'hosts': '1.2.3.4 someswitch1', 'hostname': 'someswitch1.bogus.com', 'port': "
        "1234, 'user': 'carl', 'address_family': None, 'bind_address': None, 'connect_timeout': "
        f"None, 'identities_only': 'yes', 'identity_file': '{os.path.expanduser('~')}/.ssh/mysshkey', "
        "'keyboard_interactive': None, 'password_authentication': None, 'preferred_authentication':"
        " None}"
    )


@pytest.mark.skipif(
    sys.platform.startswith("win"), reason="not currently testing ssh file resolution on windows"
)
def test_host_lookup_exact_host_in_list():
    ssh_conf = SSHConfig(f"{TEST_DATA_DIR}/files/_ssh_config")
    host = ssh_conf.lookup("someswitch1")
    assert repr(host) == (
        "Host {'hosts': '1.2.3.4 someswitch1', 'hostname': 'someswitch1.bogus.com', 'port': "
        "1234, 'user': 'carl', 'address_family': None, 'bind_address': None, 'connect_timeout': "
        f"None, 'identities_only': 'yes', 'identity_file': '{os.path.expanduser('~')}/.ssh/mysshkey', "
        "'keyboard_interactive': None, 'password_authentication': None, 'preferred_authentication':"
        " None}"
    )


@pytest.mark.skipif(
    sys.platform.startswith("win"), reason="not currently testing ssh file resolution on windows"
)
def test_host_lookup_host_fuzzy():
    ssh_conf = SSHConfig(f"{TEST_DATA_DIR}/files/_ssh_config")
    host = ssh_conf.lookup("someswitch2")
    assert repr(host) == (
        "Host {'hosts': 'someswitch?', 'hostname': 'someswitch1.bogus.com', 'port': "
        "1234, 'user': 'notcarl', 'address_family': None, 'bind_address': None, 'connect_timeout': "
        f"None, 'identities_only': 'yes', 'identity_file': '{os.path.expanduser('~')}/.ssh/mysshkey', "
        "'keyboard_interactive': None, 'password_authentication': None, 'preferred_authentication': "
        "None}"
    )


@pytest.mark.skipif(
    sys.platform.startswith("win"), reason="not currently testing ssh file resolution on windows"
)
def test_host_lookup_host_fuzzy_multi_match():
    ssh_conf = SSHConfig(f"{TEST_DATA_DIR}/files/_ssh_config")
    host = ssh_conf.lookup("someswitch9999")
    assert repr(host) == (
        "Host {'hosts': 'someswitch?', 'hostname': 'someswitch1.bogus.com', 'port': "
        "1234, 'user': 'notcarl', 'address_family': None, 'bind_address': None, 'connect_timeout': "
        f"None, 'identities_only': 'yes', 'identity_file': '{os.path.expanduser('~')}/.ssh/mysshkey', "
        "'keyboard_interactive': None, 'password_authentication': None, 'preferred_authentication': "
        "None}"
    )


@pytest.mark.skipif(
    sys.platform.startswith("win"), reason="not currently testing ssh file resolution on windows"
)
def test_host_lookup_merged_data():
    ssh_conf = SSHConfig(f"{TEST_DATA_DIR}/files/_ssh_config")
    host = ssh_conf.lookup("scrapli")
    assert repr(host) == (
        "Host {'hosts': 'scrapli', 'hostname': None, 'port': None, 'user': 'scrapli', 'address_family': None, "
        "'bind_address': None, 'connect_timeout': None, 'identities_only': None, 'identity_file': "
        f"'{os.path.expanduser('~')}/.ssh/lastresortkey', 'keyboard_interactive': None, 'password_authentication': None, "
        "'preferred_authentication': None}"
    )


def test_init_ssh_known_hosts_file_explicit():
    known_hosts = SSHKnownHosts(f"{TEST_DATA_DIR}/files/_ssh_known_hosts")
    with open(f"{TEST_DATA_DIR}/files/_ssh_known_hosts", "r") as f:
        ssh_known_hosts = f.read()
    assert known_hosts.ssh_known_hosts == ssh_known_hosts


def test_init_ssh_known_hosts_file_no_config_file(fs):
    known_hosts = SSHKnownHosts("")
    assert known_hosts.hosts == {}


def test_init_ssh_known_hosts_file_no_hosts():
    known_hosts = SSHKnownHosts(f"{TEST_DATA_DIR}/files/__init__.py")
    assert known_hosts.hosts == {}



"""scrapli network ssh client library"""
import logging
from logging import getLogger
from typing import Optional, Tuple

from scrapli.driver import AsyncScrape, Scrape
from scrapli.factory import AsyncScrapli, Scrapli

__version__ = "2020.08.28"
__all__ = ("AsyncScrape", "Scrape", "AsyncScrapli", "Scrapli")


class DuplicateFilter(logging.Filter):
    def __init__(self) -> None:
        """
        Remove duplicates from log

        Thank you to our lord and savior, StackOverflow... src:
        https://stackoverflow.com/questions/44691558/suppress-multiple-messages-with-same-content- \
        in-python-logging-module-aka-log-co

        """
        super().__init__()
        self.last_log: Optional[Tuple[str, int, str]] = None

    def filter(self, record: logging.LogRecord) -> bool:
        """
        Filter duplicate entries in logs

        Fields to compare to previous log entry if these fields match; skip log entry

        Args:
            record: log record to check

        Returns:
            bool: filter or not

        Raises:
            N/A  # noqa

        """
        current_log = (record.module, record.levelno, record.msg)
        if current_log != getattr(self, "last_log", None):
            self.last_log = current_log
            return True
        return False


SCRAPLI_LOG = getLogger("scrapli")
SCRAPLI_LOG.addFilter(DuplicateFilter())



import pytest


@pytest.mark.asyncio
async def test_nxos_open_close(async_cisco_nxos_conn):
    await async_cisco_nxos_conn.open()
    assert await async_cisco_nxos_conn.get_prompt() == "switch#"
    # "more" will show up if we haven't sent terminal length 0 to the mock nxos server
    response = await async_cisco_nxos_conn.send_command(command="show version")
    assert "more" not in response.result.lower()
    await async_cisco_nxos_conn.close()



import sys

import pytest

from scrapli.exceptions import ScrapliTimeout
from scrapli.transport.socket import Socket


@pytest.mark.skipif(sys.platform.startswith("win"), reason="not testing windows")
def test_socket_open_success():
    sock = Socket("localhost", 22, 1)
    sock.socket_open()
    assert sock.socket_isalive() is True


@pytest.mark.skipif(sys.platform.startswith("win"), reason="not testing windows")
def test_socket_open_failure_connection_refused():
    sock = Socket("localhost", 2222, 1)
    with pytest.raises(ConnectionRefusedError) as e:
        sock.socket_open()
    assert str(e.value) == "Connection refused trying to open socket to localhost on port 2222"


@pytest.mark.skipif(sys.platform.startswith("win"), reason="not testing windows")
def test_socket_open_failure_timeout():
    sock = Socket("240.0.0.1", 22, 0.1)
    with pytest.raises(ScrapliTimeout) as e:
        sock.socket_open()
    assert str(e.value) == "Timed out trying to open socket to 240.0.0.1 on port 22"


@pytest.mark.skipif(sys.platform.startswith("win"), reason="not testing windows")
def test_socket_close_success():
    sock = Socket("localhost", 22, 1)
    sock.socket_open()
    assert sock.socket_isalive() is True
    sock.socket_close()
    assert sock.socket_isalive() is False


def test_socket_isalive_false():
    sock = Socket("localhost", 22, 1)
    assert sock.socket_isalive() is False


def test__str():
    sock = Socket("localhost", 22, 1)
    assert str(sock) == "Socket Object for host localhost"


def test__repr():
    sock = Socket("localhost", 22, 1)
    assert (
        repr(sock)
        == "Socket {'logger': 'scrapli.socket-localhost', 'host': 'localhost', 'port': 22, 'timeout': "
        "1, 'sock': None}"
    )


def test__bool():
    sock = Socket("localhost", 22, 1)
    assert bool(sock) is False



"""scrapli.driver.core.cisco_iosxe.async_driver"""
from copy import deepcopy
from typing import Any, Callable, Dict, List, Optional

from scrapli.driver import AsyncNetworkDriver
from scrapli.driver.base_network_driver import PrivilegeLevel
from scrapli.driver.core.cisco_iosxe.base_driver import PRIVS


async def iosxe_on_open(conn: AsyncNetworkDriver) -> None:
    """
    AsyncIOSXEDriver default on_open callable

    Args:
        conn: NetworkDriver object

    Returns:
        N/A  # noqa: DAR202

    Raises:
        N/A
    """
    await conn.acquire_priv(desired_priv=conn.default_desired_privilege_level)
    await conn.send_command(command="terminal length 0")
    await conn.send_command(command="terminal width 512")


async def iosxe_on_close(conn: AsyncNetworkDriver) -> None:
    """
    AsyncIOSXEDriver default on_close callable

    Args:
        conn: NetworkDriver object

    Returns:
        N/A  # noqa: DAR202

    Raises:
        N/A
    """
    # write exit directly to the transport as channel would fail to find the prompt after sending
    # the exit command!
    await conn.acquire_priv(desired_priv=conn.default_desired_privilege_level)
    conn.transport.write(channel_input="exit")
    conn.transport.write(channel_input=conn.channel.comms_return_char)


class AsyncIOSXEDriver(AsyncNetworkDriver):
    def __init__(
        self,
        privilege_levels: Optional[Dict[str, PrivilegeLevel]] = None,
        default_desired_privilege_level: str = "privilege_exec",
        auth_secondary: str = "",
        on_open: Optional[Callable[..., Any]] = None,
        on_close: Optional[Callable[..., Any]] = None,
        textfsm_platform: str = "cisco_iosxe",
        genie_platform: str = "iosxe",
        failed_when_contains: Optional[List[str]] = None,
        **kwargs: Dict[str, Any],
    ):
        """
        AsyncIOSXEDriver Object

        Args:
            privilege_levels: optional user provided privilege levels, if left None will default to
                scrapli standard privilege levels
            default_desired_privilege_level: string of name of default desired priv, this is the
                priv level that is generally used to disable paging/set terminal width and things
                like that upon first login, and is also the priv level scrapli will try to acquire
                for normal "command" operations (`send_command`, `send_commands`)
            auth_secondary: password to use for secondary authentication (enable)
            on_open: callable that accepts the class instance as its only argument. this callable,
                if provided, is executed immediately after authentication is completed. Common use
                cases for this callable would be to disable paging or accept any kind of banner
                message that prompts a user upon connection
            on_close: callable that accepts the class instance as its only argument. this callable,
                if provided, is executed immediately prior to closing the underlying transport.
                Common use cases for this callable would be to save configurations prior to exiting,
                or to logout properly to free up vtys or similar.
            textfsm_platform: string name of textfsm parser platform
            genie_platform: string name of cisco genie parser platform
            failed_when_contains: List of strings that indicate a command/config has failed
            **kwargs: keyword args to pass to inherited class(es)

        Returns:
            N/A  # noqa: DAR202

        Raises:
            N/A
        """
        if privilege_levels is None:
            privilege_levels = deepcopy(PRIVS)

        if on_open is None:
            on_open = iosxe_on_open
        if on_close is None:
            on_close = iosxe_on_close

        if failed_when_contains is None:
            failed_when_contains = [
                "% Ambiguous command",
                "% Incomplete command",
                "% Invalid input detected",
                "% Unknown command",
            ]

        super().__init__(
            privilege_levels=privilege_levels,
            default_desired_privilege_level=default_desired_privilege_level,
            auth_secondary=auth_secondary,
            failed_when_contains=failed_when_contains,
            textfsm_platform=textfsm_platform,
            genie_platform=genie_platform,
            on_open=on_open,
            on_close=on_close,
            **kwargs,
        )



import sys
from datetime import datetime

import pytest

from scrapli.exceptions import ScrapliCommandFailure
from scrapli.response import MultiResponse, Response


def test_response_init():
    response = Response("localhost", "ls -al", failed_when_contains="tacocat")
    response_start_time = str(datetime.now())[:-7]
    assert response.host == "localhost"
    assert response.channel_input == "ls -al"
    assert str(response.start_time)[:-7] == response_start_time
    assert response.failed is True
    assert bool(response) is True
    assert repr(response) == "Response <Success: False>"
    assert str(response) == "Response <Success: False>"
    assert response.failed_when_contains == ["tacocat"]
    with pytest.raises(ScrapliCommandFailure):
        response.raise_for_status()


def test_multi_response():
    response1 = Response("localhost", "ls -al")
    response2 = Response("localhost", "ls -al")
    multi_response = MultiResponse([response1, response2])
    assert len(multi_response) == 2
    assert multi_response.failed is True
    assert repr(multi_response) == "MultiResponse <Success: False; Response Elements: 2>"
    assert str(multi_response) == "MultiResponse <Success: False; Response Elements: 2>"
    with pytest.raises(ScrapliCommandFailure):
        multi_response.raise_for_status()
    multi_response[0].failed = False
    multi_response[1].failed = False
    assert multi_response.failed is False
    assert multi_response.raise_for_status() is None
    assert repr(multi_response) == "MultiResponse <Success: True; Response Elements: 2>"
    assert str(multi_response) == "MultiResponse <Success: True; Response Elements: 2>"
    assert multi_response.result == "ls -al\nls -al\n"


def test_response_record_result():
    response = Response("localhost", "ls -al")
    response_end_time = str(datetime.now())[:-7]
    response_bytes = b"""
ls -al
total 264
drwxr-xr-x  34 carl  staff   1088 Jan 27 19:07 ./
drwxr-xr-x  21 carl  staff    672 Jan 25 15:56 ../
-rw-r--r--   1 carl  staff  53248 Jan 27 19:07 .coverage
drwxr-xr-x  12 carl  staff    384 Jan 27 19:13 .git/"""
    response._record_response(response_bytes)
    assert str(response.finish_time)[:-7] == response_end_time
    assert response.result == response_bytes.decode()
    assert response.failed is False


def test_response_record_result_failed_when_failed():
    response = Response("localhost", "ls -al", failed_when_contains=["!racecar!"])
    response_end_time = str(datetime.now())[:-7]
    response_bytes = b"""
ls -al
total 264
drwxr-xr-x  34 carl  staff   1088 Jan 27 19:07 ./
drwxr-xr-x  21 carl  staff    672 Jan 25 15:56 ../
-rw-r--r--   1 carl  staff  53248 Jan 27 19:07 !racecar!
drwxr-xr-x  12 carl  staff    384 Jan 27 19:13 .git/"""
    response._record_response(response_bytes)
    assert str(response.finish_time)[:-7] == response_end_time
    assert response.result == response_bytes.decode()
    assert response.failed is True


def test_response_record_result_failed_when_success():
    response = Response("localhost", "ls -al", failed_when_contains=["!racecar!"])
    response_end_time = str(datetime.now())[:-7]
    response_bytes = b"""
ls -al
total 264
drwxr-xr-x  34 carl  staff   1088 Jan 27 19:07 ./
drwxr-xr-x  21 carl  staff    672 Jan 25 15:56 ../
-rw-r--r--   1 carl  staff  53248 Jan 27 19:07 .coverage
drwxr-xr-x  12 carl  staff    384 Jan 27 19:13 .git/"""
    response._record_response(response_bytes)
    assert str(response.finish_time)[:-7] == response_end_time
    assert response.result == response_bytes.decode()
    assert response.failed is False


@pytest.mark.skipif(sys.platform.startswith("win"), reason="not supporting textfsm on windows")
@pytest.mark.parametrize(
    "parse_type",
    [
        (
            False,
            ["Internet", "172.31.254.1", "-", "0000.0c07.acfe", "ARPA", "Vlan254"],
        ),
        (
            True,
            {
                "protocol": "Internet",
                "address": "172.31.254.1",
                "age": "-",
                "mac": "0000.0c07.acfe",
                "type": "ARPA",
                "interface": "Vlan254",
            },
        ),
    ],
    ids=["to_dict_false", "to_dict_true"],
)
def test_response_parse_textfsm(parse_type):
    to_dict = parse_type[0]
    expected_result = parse_type[1]
    response = Response("localhost", channel_input="show ip arp", textfsm_platform="cisco_ios")
    response_bytes = b"""Protocol  Address          Age (min)  Hardware Addr   Type   Interface
Internet  172.31.254.1            -   0000.0c07.acfe  ARPA   Vlan254
Internet  172.31.254.2            -   c800.84b2.e9c2  ARPA   Vlan254
"""
    response._record_response(response_bytes)
    assert response.textfsm_parse_output(to_dict=to_dict)[0] == expected_result


@pytest.mark.skipif(sys.platform.startswith("win"), reason="not supporting textfsm on windows")
def test_response_parse_textfsm_fail():
    response = Response("localhost", channel_input="show ip arp", textfsm_platform="cisco_ios")
    response_bytes = b""
    response._record_response(response_bytes)
    assert response.textfsm_parse_output() == []


@pytest.mark.skipif(sys.platform.startswith("win"), reason="not supporting genie on windows")
def test_response_parse_genie():
    response = Response("localhost", channel_input="show ip arp", genie_platform="iosxe")
    response_bytes = b"""Protocol  Address          Age (min)  Hardware Addr   Type   Interface
Internet  172.31.254.1            -   0000.0c07.acfe  ARPA   Vlan254
Internet  172.31.254.2            -   c800.84b2.e9c2  ARPA   Vlan254
"""
    response._record_response(response_bytes)
    result = response.genie_parse_output()
    assert (
        result["interfaces"]["Vlan254"]["ipv4"]["neighbors"]["172.31.254.1"]["ip"] == "172.31.254.1"
    )


@pytest.mark.skipif(sys.platform.startswith("win"), reason="not supporting genie on windows")
def test_response_parse_genie_fail():
    response = Response("localhost", channel_input="show ip arp", genie_platform="iosxe")
    response_bytes = b""
    response._record_response(response_bytes)
    assert response.genie_parse_output() == []



import os
import sys
from pathlib import Path

import pytest

import scrapli
from scrapli.exceptions import ConnectionNotOpened, ScrapliAuthenticationFailed
from scrapli.transport import SystemSSHTransport

TEST_DATA_DIR = f"{Path(scrapli.__file__).parents[1]}/tests/test_data"


def test_str():
    conn = SystemSSHTransport("localhost")
    assert str(conn) == "Transport Object for host localhost"


def test_repr():
    conn = SystemSSHTransport("localhost")
    assert (
        repr(conn)
        == "Transport {'logger': 'scrapli.transport-localhost', 'host': 'localhost', 'port': 22, 'timeout_socket': 5, "
        "'timeout_transport': 5, 'timeout_exit': True, 'keepalive': False, 'keepalive_interval': 30, "
        "'keepalive_type': '', 'keepalive_pattern': '\\x05', 'session_lock': False, 'auth_username': '', "
        "'auth_private_key': '', 'auth_private_key_passphrase': '********', 'auth_password': '********', "
        "'auth_strict_key': True, 'auth_bypass': False, '_timeout_ops': 10, '_comms_prompt_pattern': '^["
        "a-z0-9.\\\\-@()/:]{1,32}[#>$]$', '_comms_return_char': '\\n', '_comms_ansi': False, 'ssh_config_file': "
        "'', 'ssh_known_hosts_file': '', 'lib_auth_exception': <class "
        "'scrapli.exceptions.ScrapliAuthenticationFailed'>, '_isauthenticated': False, 'transport_options': {}, "
        "'open_cmd': ['ssh', 'localhost', '-p', '22', '-o', 'ConnectTimeout=5', '-o', 'ServerAliveInterval=5', "
        "'-o', 'StrictHostKeyChecking=yes', '-F', '/dev/null']}"
    )


@pytest.mark.skipif(sys.platform.startswith("win"), reason="systemssh not supported on windows")
def test_creation():
    conn = SystemSSHTransport("localhost")
    assert conn.host == "localhost"
    assert conn.port == 22
    assert conn._isauthenticated is False


@pytest.mark.skipif(sys.platform.startswith("win"), reason="systemssh not supported on windows")
def test_build_open_cmd():
    conn = SystemSSHTransport("localhost")
    assert conn.open_cmd == [
        "ssh",
        "localhost",
        "-p",
        "22",
        "-o",
        "ConnectTimeout=5",
        "-o",
        "ServerAliveInterval=5",
        "-o",
        "StrictHostKeyChecking=yes",
        "-F",
        "/dev/null",
    ]


@pytest.mark.skipif(sys.platform.startswith("win"), reason="systemssh not supported on windows")
@pytest.mark.parametrize(
    "user_options",
    [
        ["oKexAlgorithms=+diffie-hellman-group1-sha1"],
        "oKexAlgorithms=+diffie-hellman-group1-sha1",
    ],
    ids=[
        "user options list",
        "user options string",
    ],
)
def test_build_open_cmd_user_options(user_options):
    conn = SystemSSHTransport("localhost", transport_options={"open_cmd": user_options})
    assert conn.open_cmd == [
        "ssh",
        "localhost",
        "-p",
        "22",
        "-o",
        "ConnectTimeout=5",
        "-o",
        "ServerAliveInterval=5",
        "-o",
        "StrictHostKeyChecking=yes",
        "-F",
        "/dev/null",
        "oKexAlgorithms=+diffie-hellman-group1-sha1",
    ]


@pytest.mark.skipif(sys.platform.startswith("win"), reason="systemssh not supported on windows")
@pytest.mark.parametrize(
    "eof_msg",
    [
        (
            b"Host key verification failed",
            "Host key verification failed for host localhost",
        ),
        (
            b"Operation timed out",
            "Timed out connecting to host localhost",
        ),
        (
            b"Connection timed out",
            "Timed out connecting to host localhost",
        ),
        (
            b"No route to host",
            "No route to host localhost",
        ),
        (
            b"no matching key exchange found.",
            "No matching key exchange found for host localhost",
        ),
        (
            b"no matching key exchange found. Their offer: diffie-hellman-group-exchange-sha1,diffie-hellman-group14-sha1",
            "No matching key exchange found for host localhost, their offer: diffie-hellman-group-exchange-sha1,diffie-hellman-group14-sha1",
        ),
        (
            b"no matching cipher found",
            "No matching cipher found for host localhost",
        ),
        (
            b"no matching cipher found, their offer: aes128-cbc,aes256-cbc",
            "No matching cipher found for host localhost, their offer: aes128-cbc,aes256-cbc",
        ),
        (
            b"command-line: line 0: Bad configuration option: ciphers+",
            "Bad SSH configuration option(s) for host localhost, bad option(s): ciphers+",
        ),
        (
            b"WARNING: UNPROTECTED PRIVATE KEY FILE!",
            # note: empty quotes in the middle is where private key filename would be
            "Permissions for private key `` are too open, authentication failed!",
        ),
        (
            b"Could not resolve hostname BLAH: No address associated with hostname",
            # note: empty quotes in the middle is where private key filename would be
            "Could not resolve address for host `localhost`",
        ),
    ],
    ids=[
        "host key verification",
        "operation time out",
        "connection time out",
        "no route to host",
        "no matching key exchange",
        "no matching key exchange found key exchange",
        "no matching cipher",
        "no matching cipher found ciphers",
        "bad configuration option",
        "unprotected key",
        "could not resolve host",
    ],
)
def test_ssh_message_handler(eof_msg):
    conn = SystemSSHTransport("localhost")
    error_msg = eof_msg[0]
    expected_msg = eof_msg[1]
    with pytest.raises(ScrapliAuthenticationFailed) as exc:
        conn._ssh_message_handler(error_msg)
    assert str(exc.value) == expected_msg


@pytest.mark.skipif(sys.platform.startswith("win"), reason="systemssh not supported on windows")
def test_set_timeout():
    conn = SystemSSHTransport("localhost")
    assert conn.timeout_transport == 5
    conn.set_timeout(1000)
    assert conn.timeout_transport == 1000
    conn.timeout_transport = 9999
    conn.set_timeout()
    assert conn.timeout_transport == 9999


@pytest.mark.skipif(sys.platform.startswith("win"), reason="systemssh not supported on windows")
@pytest.mark.parametrize(
    "method_name",
    ["read", "write"],
    ids=["read", "write"],
)
def test_requires_open(method_name):
    conn = SystemSSHTransport("localhost")
    method = getattr(conn, method_name)
    with pytest.raises(ConnectionNotOpened):
        if method_name == "write":
            method("blah")
        else:
            method()



import logging
import time
from pathlib import Path

from device_info import iosxr_device

from scrapli.driver.core import IOSXRDriver

logging.basicConfig(
    filename=f"{Path(__file__).resolve().parents[0]}/iosxr_driver.log", level=logging.DEBUG
)
logger = logging.getLogger("scrapli")

conn = IOSXRDriver(**iosxr_device)
conn.open()

print("***** Get Prompt:")
prompt = conn.get_prompt()
print(prompt)

print("***** Show run | i hostname:")
result = conn.send_command("show run | i hostname")
print(result, result.result)

print("***** Show run:")
result = conn.send_command("show run")
print(result, result.result)

if iosxr_device["keepalive"]:
    print("***** Waiting for keepalive....")
    time.sleep(5)

conn.close()



"""scrapli.driver"""
from scrapli.driver.async_driver import AsyncScrape
from scrapli.driver.async_generic_driver import AsyncGenericDriver
from scrapli.driver.async_network_driver import AsyncNetworkDriver
from scrapli.driver.driver import Scrape
from scrapli.driver.generic_driver import GenericDriver
from scrapli.driver.network_driver import NetworkDriver

__all__ = (
    "AsyncScrape",
    "Scrape",
    "AsyncGenericDriver",
    "GenericDriver",
    "AsyncNetworkDriver",
    "NetworkDriver",
)



import pytest

from scrapli.transport import Transport

from ...test_data.unit_test_cases import TEST_CASES


def test_init(sync_cisco_iosxe_conn):
    """Test that all arguments get properly passed from driver/transport to channel on init"""
    assert isinstance(sync_cisco_iosxe_conn.channel.transport, Transport)
    assert (
        sync_cisco_iosxe_conn.channel.comms_prompt_pattern
        == sync_cisco_iosxe_conn._initialization_args["comms_prompt_pattern"]
    )
    assert (
        sync_cisco_iosxe_conn.channel.comms_return_char
        == sync_cisco_iosxe_conn._initialization_args["comms_return_char"]
    )
    assert (
        sync_cisco_iosxe_conn.channel.comms_ansi
        == sync_cisco_iosxe_conn._initialization_args["comms_ansi"]
    )
    # auto expand is not set in initialization args at this point (may 2020)
    assert (
        sync_cisco_iosxe_conn.channel.comms_auto_expand
        == sync_cisco_iosxe_conn._initialization_args.get("comms_auto_expand", False)
    )
    assert (
        sync_cisco_iosxe_conn.channel.timeout_ops
        == sync_cisco_iosxe_conn._initialization_args["timeout_ops"]
    )


def test_read_chunk(sync_cisco_iosxe_conn):
    sync_cisco_iosxe_conn.open()
    chunk = "this is some input for read chunk to read"
    sync_cisco_iosxe_conn.transport.write(chunk)
    sync_cisco_iosxe_conn.transport.write(sync_cisco_iosxe_conn.channel.comms_return_char)
    # keep reading chunks until we have all the input data
    read_chunk = b""
    while chunk.encode() not in read_chunk:
        read_chunk += sync_cisco_iosxe_conn.channel._read_chunk()


def test_read_until_input(sync_cisco_iosxe_conn):
    sync_cisco_iosxe_conn.open()
    chunk = "this is some input for read chunk to read"
    sync_cisco_iosxe_conn.transport.write(chunk)
    sync_cisco_iosxe_conn.transport.write(sync_cisco_iosxe_conn.channel.comms_return_char)
    read_chunk = sync_cisco_iosxe_conn.channel._read_until_input(channel_input=chunk.encode())
    assert chunk.encode() in read_chunk


def test_read_until_prompt(sync_cisco_iosxe_conn):
    expected_prompt = TEST_CASES["cisco_iosxe"]["test_read_until_prompt"]["expected_prompt"]

    sync_cisco_iosxe_conn.open()
    chunk = "this is some input for read chunk to read"
    sync_cisco_iosxe_conn.transport.write(chunk)
    sync_cisco_iosxe_conn.transport.write(sync_cisco_iosxe_conn.channel.comms_return_char)
    read_chunk = sync_cisco_iosxe_conn.channel._read_until_input(channel_input=chunk.encode())
    found_prompt = sync_cisco_iosxe_conn.get_prompt()
    assert chunk.encode() in read_chunk
    assert found_prompt == expected_prompt


def test_read_until_prompt_or_time(sync_cisco_iosxe_conn):
    sync_cisco_iosxe_conn.open()
    chunk = "this is some input for read chunk to read"
    sync_cisco_iosxe_conn.transport.write(chunk)
    sync_cisco_iosxe_conn.transport.write(sync_cisco_iosxe_conn.channel.comms_return_char)
    read_chunk = sync_cisco_iosxe_conn.channel._read_until_prompt_or_time(
        channel_outputs=[b"chunk to read"]
    )
    assert chunk.encode() in read_chunk


def test_get_prompt(sync_cisco_iosxe_conn):
    expected_prompt = TEST_CASES["cisco_iosxe"]["test_get_prompt"]["privilege_exec"]
    sync_cisco_iosxe_conn.open()
    found_prompt = sync_cisco_iosxe_conn.get_prompt()
    assert found_prompt == expected_prompt


@pytest.mark.parametrize(
    "strip_prompt",
    [True, False],
    ids=["strip_prompt", "no_strip_prompt"],
)
def test_send_input(sync_cisco_iosxe_conn, strip_prompt):
    expected_raw = TEST_CASES["cisco_iosxe"]["test_send_input"]["raw_result"]
    expected_processed = (
        TEST_CASES["cisco_iosxe"]["test_send_input"]["processed_result"]["strip"]
        if strip_prompt
        else TEST_CASES["cisco_iosxe"]["test_send_input"]["processed_result"]["no_strip"]
    )
    sync_cisco_iosxe_conn.open()
    raw_result, processed_result = sync_cisco_iosxe_conn.channel.send_input(
        channel_input="show version", strip_prompt=strip_prompt
    )
    assert raw_result == expected_raw.encode()
    assert processed_result == expected_processed.encode()


def test_send_inputs_interact(sync_cisco_iosxe_conn):
    expected_raw = TEST_CASES["cisco_iosxe"]["test_send_inputs_interact"]["raw_result"]
    expected_processed = TEST_CASES["cisco_iosxe"]["test_send_inputs_interact"]["processed_result"]
    interact_events = TEST_CASES["cisco_iosxe"]["test_send_inputs_interact"]["interact_events"]
    sync_cisco_iosxe_conn.open()
    raw_result, processed_result = sync_cisco_iosxe_conn.channel.send_inputs_interact(
        interact_events=interact_events
    )
    assert raw_result == expected_raw.encode()
    assert processed_result == expected_processed.encode()


@pytest.mark.parametrize(
    "strip_prompt",
    [True, False],
    ids=["strip_prompt", "no_strip_prompt"],
)
def test_send_input_and_read(sync_cisco_iosxe_conn, strip_prompt):
    expected_raw = TEST_CASES["cisco_iosxe"]["test_send_input"]["raw_result"]
    expected_processed = (
        TEST_CASES["cisco_iosxe"]["test_send_input"]["processed_result"]["strip"]
        if strip_prompt
        else TEST_CASES["cisco_iosxe"]["test_send_input"]["processed_result"]["no_strip"]
    )
    sync_cisco_iosxe_conn.open()
    raw_result, processed_result = sync_cisco_iosxe_conn.channel.send_input_and_read(
        channel_input="show version", strip_prompt=strip_prompt
    )
    assert raw_result == expected_raw.encode()
    assert processed_result == expected_processed.encode()



import asyncio
import time
from concurrent.futures import ThreadPoolExecutor
from pathlib import Path

import pytest

import scrapli
from scrapli.driver import AsyncGenericDriver, GenericDriver
from scrapli.driver.core import (
    AsyncIOSXEDriver,
    AsyncJunosDriver,
    AsyncNXOSDriver,
    IOSXEDriver,
    JunosDriver,
    NXOSDriver,
)

from ..mock_devices.run import run
from ..test_data.devices import DEVICES, ENCRYPTED_PRIVATE_KEY, PRIVATE_KEY

TEST_DATA_DIR = f"{Path(scrapli.__file__).parents[1]}/tests/test_data"
SERVER_KEY = f"{TEST_DATA_DIR}/files/vrnetlab_key"

SYNC_DRIVERS = {
    "cisco_iosxe": IOSXEDriver,
    "cisco_nxos": NXOSDriver,
    "juniper_junos": JunosDriver,
    "generic_driver": GenericDriver,
}
ASYNC_DRIVERS = {
    "cisco_iosxe": AsyncIOSXEDriver,
    "cisco_nxos": AsyncNXOSDriver,
    "juniper_junos": AsyncJunosDriver,
    "generic_driver": AsyncGenericDriver,
}


@pytest.fixture(scope="session", autouse=True)
def mock_ssh_server_cisco_iosxe():
    with ThreadPoolExecutor(max_workers=1) as pool:
        loop = asyncio.new_event_loop()
        pool.submit(run, SERVER_KEY, "cisco_iosxe", loop)
        # short sleep -- seems the servers want a quick second to "wake up"
        time.sleep(0.25)
        # yield to let all the tests run, then we can deal w/ cleaning up the thread/loop
        yield
        loop.call_soon_threadsafe(loop.stop)


@pytest.fixture(scope="session", autouse=True)
def mock_ssh_server_cisco_nxos():
    with ThreadPoolExecutor(max_workers=1) as pool:
        loop = asyncio.new_event_loop()
        pool.submit(run, SERVER_KEY, "cisco_nxos", loop)
        # short sleep -- seems the servers want a quick second to "wake up"
        time.sleep(0.25)
        # yield to let all the tests run, then we can deal w/ cleaning up the thread/loop
        yield
        loop.call_soon_threadsafe(loop.stop)


@pytest.fixture(scope="session", autouse=True)
def mock_ssh_server_juniper_junos():
    with ThreadPoolExecutor(max_workers=1) as pool:
        loop = asyncio.new_event_loop()
        pool.submit(run, SERVER_KEY, "juniper_junos", loop)
        # short sleep -- seems the servers want a quick second to "wake up"
        time.sleep(0.25)
        # yield to let all the tests run, then we can deal w/ cleaning up the thread/loop
        yield
        loop.call_soon_threadsafe(loop.stop)


@pytest.fixture(scope="function")
def sync_generic_driver_conn():
    device = DEVICES["mock_cisco_iosxe"].copy()
    device.pop("auth_secondary")
    driver = SYNC_DRIVERS["generic_driver"]
    conn = driver(**device)
    yield conn
    if conn.isalive():
        conn.close()


@pytest.fixture(scope="function")
async def async_generic_driver_conn():
    device = DEVICES["mock_cisco_iosxe"].copy()
    device.pop("auth_secondary")
    driver = ASYNC_DRIVERS["generic_driver"]
    conn = driver(transport="asyncssh", **device)
    yield conn
    if conn.isalive():
        await conn.close()


@pytest.fixture(scope="function", params=["password", "key"])
def sync_conn_auth_type(request):
    # fixture to ensue we use password and key based auth with SystemSSHTransport
    yield request.param


@pytest.fixture(scope="function")
def sync_cisco_iosxe_conn(sync_conn_auth_type):
    device = DEVICES["mock_cisco_iosxe"].copy()
    if sync_conn_auth_type == "key":
        device.pop("auth_password")
        device["auth_private_key"] = PRIVATE_KEY
    driver = SYNC_DRIVERS["cisco_iosxe"]
    conn = driver(**device)
    yield conn
    if conn.isalive():
        conn.close()


@pytest.fixture(scope="function")
def sync_cisco_iosxe_conn_encrypted_key():
    device = DEVICES["mock_cisco_iosxe"].copy()
    device.pop("auth_password")
    device["auth_private_key"] = ENCRYPTED_PRIVATE_KEY
    driver = SYNC_DRIVERS["cisco_iosxe"]
    conn = driver(**device)
    yield conn
    if conn.isalive():
        conn.close()


@pytest.fixture(scope="function")
async def async_cisco_iosxe_conn():
    device = DEVICES["mock_cisco_iosxe"]
    driver = ASYNC_DRIVERS["cisco_iosxe"]
    conn = driver(transport="asyncssh", **device)
    yield conn
    if conn.isalive():
        await conn.close()


@pytest.fixture(scope="function")
def sync_cisco_nxos_conn(sync_conn_auth_type):
    device = DEVICES["mock_cisco_nxos"].copy()
    if sync_conn_auth_type == "key":
        device.pop("auth_password")
        device["auth_private_key"] = PRIVATE_KEY
    driver = SYNC_DRIVERS["cisco_nxos"]
    conn = driver(**device)
    yield conn
    if conn.isalive():
        conn.close()


@pytest.fixture(scope="function")
async def async_cisco_nxos_conn():
    device = DEVICES["mock_cisco_nxos"]
    driver = ASYNC_DRIVERS["cisco_nxos"]
    conn = driver(transport="asyncssh", **device)
    yield conn
    if conn.isalive():
        await conn.close()


@pytest.fixture(scope="function")
def sync_juniper_junos_conn(sync_conn_auth_type):
    device = DEVICES["mock_juniper_junos"].copy()
    if sync_conn_auth_type == "key":
        device.pop("auth_password")
        device["auth_private_key"] = PRIVATE_KEY
    driver = SYNC_DRIVERS["juniper_junos"]
    conn = driver(transport="ssh2", **device)
    yield conn
    if conn.isalive():
        conn.close()


@pytest.fixture(scope="function")
async def async_juniper_junos_conn():
    device = DEVICES["mock_juniper_junos"]
    driver = ASYNC_DRIVERS["juniper_junos"]
    conn = driver(transport="asyncssh", **device)
    yield conn
    if conn.isalive():
        await conn.close()



def test_configuration_exclusive(junos_conn):
    junos_conn.acquire_priv("configuration_exclusive")
    result = junos_conn.send_configs(["status"])
    junos_conn.close()
    assert "exclusive" in result[0].result


def test_configuration_private(junos_conn):
    junos_conn.acquire_priv("configuration_private")
    result = junos_conn.send_configs(["status"])
    junos_conn.close()
    assert "private" in result[0].result
